--  _    _          ___           __ _     (R)
-- | |  (_)_ _____ / __|___ _ _  / _(_)__ _
-- | |__| \ V / -_) (__/ _ \ ' \|  _| / _` |
-- |____|_|\_/\___|\___\___/_||_|_| |_\__, |
--                                    |___/
-- Copyright (c) 2009-2022 LiveConfig GmbH.
-- ---------------------------------------------------------------------------
-- common/lua/apache.lua
-- Lua module to manage Apache httpd web server
-- ---------------------------------------------------------------------------
-- This is the "driver" module for Apache httpd web server.
-- It must be loaded by the command
--   LC.web.load("apache")
-- Usually, this should happen at liveconfig.lua (or, if you have a customized
-- module, at custom.lua)
-- ---------------------------------------------------------------------------

-- Imported functions
local LC = LC
local LCS = LCS -- import liveconfig global storage
local assert = assert
local type = type
local io = io
local os = os
local ipairs = ipairs
local pairs = pairs
local string = string
local table = table     -- for table.sort
local tonumber = tonumber

-- Module declaration
module("apache")

-- Set configuration mode for virtual hosts
-- Possible values:
-- 'virtualhost': generate a <VirtualHost> section for each different
--                domain configuration (better compatibility)
-- 'rewrite': generate only one <VirtualHost> per subscription, configure all
--            additional domains using RewriteRules (better performance when
--            managing >1000 subscriptions on one host)
-- Default is "virtualhost", can be overridden in "custom.lua"
if CONFIGMODE == nil then
  CONFIGMODE = 'virtualhost'
end

-- optionally allow usage of option "FollowSymLinks"
-- NOTE: THIS IS A GREAT SECURITY RISK! USE THIS ONLY WHEN YOU KNOW *EXACTLY*
--       WHAT YOU'RE DOING!
--       NEVER, NEVER, NEVER USE THIS OPTION IN A SHARED HOSTING ENVIRONMENT,
--       AS THIS EFFECTIVELY ENABLES USERS TO ACCESS FILES IN EVERY OTHER
--       CUSTOMERS DIRECTORIES!
if FOLLOWSYMLINKS == nil then
  FOLLOWSYMLINKS = false
end

-- reload interval of Apache service
-- (earliest after RELOAD_MIN seconds, latest after RELOAD_MAX seconds)
if RELOAD_MIN == nil then
  RELOAD_MIN = 10
end
if RELOAD_MAX == nil then
  RELOAD_MAX = 60
end

-- custom options for SSLProtocols setting (https://httpd.apache.org/docs/2.4/mod/mod_ssl.html#sslprotocol)
if SSL_PROTOCOLS == nil then
  SSL_PROTOCOLS = "ALL -SSLv2 -SSLv3"
end

if SSL_PROTOCOLS_STRONG == nil then
  SSL_PROTOCOLS_STRONG = "ALL -SSLv2 -SSLv3 -TLSv1 -TLSv1.1"
end

-- default SSL vhost configuration overrides
if SSLDEFAULT == nil then
  SSLDEFAULT = { }
end

-- If RESTRICT_FILEINFO is a non-empty table, only the explicitly defined Apache directives
-- are allowed in .htaccess files.
-- If RESTRICT_FILEINFO is an empty table (=default), full "FileInfo" is allowed.
-- See https://httpd.apache.org/docs/2.4/mod/overrides.html for all overrides.
if RESTRICT_FILEINFO == nil then
  RESTRICT_FILEINFO = {
  --[[
    "AcceptPathInfo",
    "Action",
    "AddCharset",
    "AddDefaultCharset",
    "AddEncoding",
    "AddHandler",
    "AddInputFilter",
    "AddLanguage",
    "AddOutputFilter",
    "AddOutputFilterByType",
    "AddType",
    "BrowserMatch",
    "BrowserMatchNoCase",
    "CGIMapExtension",
    "CGIVar",
    "CharsetDefault",
    "CharsetOptions",
    "CharsetSourceEnc",
    "CookieDomain",
    "CookieExpires",
    "CookieName",
    "CookieStyle",
    "CookieTracking",
    "DefaultLanguage",
    "DefaultType",
    "EnableMMAP",
    "EnableSendfile",
    "ErrorDocument",
    "FileETag",
    "ForceLanguagePriority",
    "ForceType",
    "Header",
    "LanguagePriority",
    "MultiviewsMatch",
    "PassEnv",
    "QualifyRedirectURL",
    "Redirect",
    "RedirectMatch",
    "RedirectPermanent",
    "RedirectTemp",
    "RemoveCharset",
    "RemoveEncoding",
    "RemoveHandler",
    "RemoveInputFilter",
    "RemoveLanguage",
    "RemoveOutputFilter",
    "RemoveType",
    "RequestHeader",
    "RewriteBase",
    "RewriteCond",
    "RewriteEngine",
    "RewriteOptions",
    "RewriteRule",
    "ScriptInterpreterSource",
    "SetEnv",
    "SetEnvIf",
    "SetEnvIfExpr",
    "SetEnvIfNoCase",
    "SetInputFilter",
    "SetOutputFilter",
    "Substitute",
    "SubstituteInheritBefore",
    "SubstituteMaxLineLength",
    "UnsetEnv"
  ]]--
  }
end


-- Exported functions
--   detect()
--   install()
--   uninstall()
--   configure()
--   configureVHost()
--   reload()
--   restart()
--   deleteAccount()

-- local table for cached evaluations of php.ini expressions (E_ALL etc.)
local phpEval = {}

-- ---------------------------------------------------------------------------
-- escape()
--
-- Escape quotes in a string to make them safe for config files
-- ---------------------------------------------------------------------------
local function escape(s)
  s=string.gsub(s, '\\', '\\\\')
  s=string.gsub(s, '"', '\\"')
  return(s)
end

-- ---------------------------------------------------------------------------
-- getBinaryVersion()
--
-- Return the binary version number from Apache httpd
-- ---------------------------------------------------------------------------
local function getBinaryVersion(bin)
  local handle = io.popen(bin .. " -v", "r")
  local ret = handle:read("*a")
  handle:close()
  local v = string.match(ret, "Apache/(%d+%.%d+[%.%d+]*)")
  return v
end

-- ---------------------------------------------------------------------------
-- check_modules()
--
-- Get list of enabled Apache modules
-- ---------------------------------------------------------------------------
local function check_modules(apachectl, args)
  if not LC.fs.is_file(apachectl) then return nil end
  local fh = assert(io.popen(apachectl .. " " .. args .. " 2>&1", "r"))
  local list = {}
  while true do
    local line = fh:read()
    if line == nil then break end
    local m = string.match(line, "([%w_]+)_module")
    if m then
      list[#list+1] = m
    end
  end
  fh:close()
  return list
end

-- ---------------------------------------------------------------------------
-- check_http2()
--
-- Check if Apache version supports HTTP/2
-- ---------------------------------------------------------------------------
local function check_http2(data)
  local v_major, v_minor, v_patch = string.match(data.binversion, "^(%d+)%.(%d+)%.(%d+)")
  v_major = tonumber(v_major)
  v_minor = tonumber(v_minor)
  v_patch = tonumber(v_patch)
  if v_major > 2 or (v_major == 2 and (v_minor > 4 or (v_minor == 4 and v_patch >= 17))) then
    data.has_http2 = true
  else
    data.has_http2 = false
  end
end

-- ---------------------------------------------------------------------------
-- detect()
--
-- Detect if an Apache httpd is installed. If yes, return a table containing
-- all relevant informations:
--   type          => "apache"
--   binname       => file name of the Apache httpd binary (eg. "/usr/sbin/apache2")
--   binversion    => binary version (eg. "2.2.3")
--   pkgname       => name of the package (see below)
--   pkgversion    => package version number (eg. "2.2.3-4+etch11")
--   configpath    => Main configuration path for Apache
--   configfile    => LiveConfigs' configuration file (snippet)
--   statusfile    => LiveConfigs' status file (containing config version etc.)
--   start_cmd     => command to start Apache httpd
--   stop_cmd      => command to stop Apache httpd
--   reload_cmd    => command to re-load configuration
--   restart_cmd   => command to re-start Apache process
--   available_dir => path for all available vhost configuration files
--   enabled_dir   => path for enabled vhost configuration files
--   enable_cmd    => command to enable a vhost configuration file (eg. "/usr/sbin/a2ensite")
--   disable_cmd   => command to disable a vhost configuration file (eg. "/usr/sbin/a2dissite")
--   httpd_user    => user running Apache httpd
--   httpd_group   => group running Apache httpd
--   has_http2     => true if Apache should support HTTP/2 (does NOT check whether "mod_http2" is enabled!)
--
-- If pkgversion is 'nil', then Apache httpd was most propably not installed
-- as native distribution package, but compiled/installed manually.
-- ---------------------------------------------------------------------------
function detect()

  -- first check for distribution-specific packages
  if LC.distribution.family == "Debian" then
    -- Debian/Ubuntu
    -- Package is named "apache2", but is only a meta-package; the real
    -- installation is made via "apache2-mpm-[...]", so we look for a list
    -- of packages.
    local pkg, v = LC.distribution.hasPackage( 'apache2', 'apache2-mpm-prefork', 'apache2-mpm-worker', 'apache2-mpm-event', 'apache2-mpm-perchild', 'apache2-mpm-itk')
    if pkg ~= nil then
      LC.log.print(LC.log.DEBUG, "Found package '", pkg, "' (Version ", v, ")")
      local bin = "/usr/sbin/apache2"
      -- get binary version
      local bv = getBinaryVersion(bin)
      LC.log.print(LC.log.DEBUG, "binver ", bv)
      if bv ~= nil then
        -- ok, we have all informations. return data table:
        local data = {
            ["type"]          = "apache",
            ["binname"]       = bin,
            ["binversion"]    = bv,
            ["pkgname"]       = pkg,
            ["pkgversion"]    = v,
            ["configpath"]    = "/etc/apache2",
            ["configfile"]    = "/etc/apache2/conf.d/liveconfig.conf",
            ["statusfile"]    = "/etc/apache2/liveconfig.status",
            ["start_cmd"]     = "/etc/init.d/apache2 start",
            ["stop_cmd"]      = "/etc/init.d/apache2 stop",
            ["reload_cmd"]    = "/etc/init.d/apache2 reload",
            ["restart_cmd"]   = "/etc/init.d/apache2 restart",
            ["available_dir"] = "/etc/apache2/sites-available",
            ["enabled_dir"]   = "/etc/apache2/sites-enabled",
            ["enable_cmd"]    = "/usr/sbin/a2ensite",
            ["disable_cmd"]   = "/usr/sbin/a2dissite",
            ["httpd_user"]    = "www-data",
            ["httpd_group"]   = "www-data",
            ["ssl_chain_dir"] = "/etc/ssl/chains",
            ["ssl_cert_dir"]  = "/etc/ssl/certs",
            ["ssl_key_dir"]   = "/etc/ssl/private",
            ["ssl_key_user"]  = "root",
            ["ssl_key_group"] = "ssl-cert",
            ["ssl_key_mode"]  = "0640",
            ["htdocs_path"]   = LC.web.HTDOCS_PATH,
            ["modules"]       = check_modules("/usr/sbin/apache2ctl", "-M"),
            ["run_dir"]       = "/var/run/apache2",
            ["reload_max"]    = RELOAD_MAX,
            ["units"]         = { ["apache"] = "apache2" }
        }
        if LC.fs.is_dir("/etc/apache2/conf-available") and LC.fs.is_dir("/etc/apache2/conf-enabled") then
          -- new config layout (Ubuntu 13.10+, Debian 8+)
          data["configfile"] = "/etc/apache2/conf-enabled/liveconfig.conf"
        end
        check_http2(data)
        return data
      end
      -- else: fall trough, to check for custom Apache installation
      LC.log.print(LC.log.DEBUG, "LC.apache.detect(): Found Apache package '", pkg, "', but no binary at ", bin)
    end
  
  elseif LC.distribution.family == "Gentoo" then
    -- Gentoo
    local pkg, v = LC.distribution.hasPackage('apache')

    if pkg ~= nil then
      LC.log.print(LC.log.DEBUG, "Found package '", pkg, "' (Version ", v, ")")
      local bin = "/usr/sbin/apache2"
      local bv = getBinaryVersion(bin)
      LC.log.print(LC.log.DEBUG, "binver ", bv)
      if bv ~= nil then
        local data = {
            ["type"]          = "apache",
            ["binname"]       = bin,
            ["binversion"]    = bv,
            ["pkgname"]       = pkg,
            ["pkgversion"]    = v,
            ["configpath"]    = "/etc/apache2",
            ["configfile"]    = "/etc/apache2/modules.d/99_liveconfig.conf",
            ["statusfile"]    = "/etc/apache2/liveconfig.status",
            ["start_cmd"]     = "/etc/init.d/apache2 start",
            ["stop_cmd"]      = "/etc/init.d/apache2 stop",
            ["reload_cmd"]    = "/etc/init.d/apache2 reload",
            ["restart_cmd"]   = "/etc/init.d/apache2 restart",
            ["available_dir"] = nil,
            ["enabled_dir"]   = "/etc/apache2/vhosts.d",
            ["enable_cmd"]    = nil,
            ["disable_cmd"]   = nil,
            ["httpd_user"]    = "apache",
            ["httpd_group"]   = "apache",
            ["ssl_chain_dir"] = "/etc/ssl/apache2",
            ["ssl_cert_dir"]  = "/etc/ssl/apache2",
            ["ssl_key_dir"]   = "/etc/ssl/apache2",
            ["ssl_key_user"]  = "root",
            ["ssl_key_group"] = "root",
            ["ssl_key_mode"]  = "0600",
            ["htdocs_path"]   = LC.web.HTDOCS_PATH,
            ["modules"]       = check_modules("/etc/init.d/apache2", "modules"),
            ["run_dir"]       = "/var/run/apache2",
            ["reload_max"]    = RELOAD_MAX,
            ["units"]         = { ["apache"] = "apache2" }
        }
        check_http2(data)
        return data
      end
      -- else: fall trough, to check for custom Apache installation
      LC.log.print(LC.log.DEBUG, "LC.apache.detect(): Found Apache package '", pkg, "', but no binary at ", bin)
    end

  elseif LC.distribution.family == "RedHat" then
    -- CentOS
    local pkg, v = LC.distribution.hasPackage('httpd')
    if pkg ~= nil then
      LC.log.print(LC.log.DEBUG, "Found package '", pkg, "' (Version ", v, ")")
      local bin = "/usr/sbin/httpd"
      local bv = getBinaryVersion(bin)
      LC.log.print(LC.log.DEBUG, "binver ", bv)
      if bv ~= nil then
        local data = {
            ["type"]          = "apache",
            ["binname"]       = bin,
            ["binversion"]    = bv,
            ["pkgname"]       = pkg,
            ["pkgversion"]    = v,
            ["configpath"]    = "/etc/httpd",
            ["configfile"]    = "/etc/httpd/conf.d/99_liveconfig.conf",
            ["statusfile"]    = "/etc/httpd/liveconfig.status",
            ["start_cmd"]     = "/sbin/service httpd start",
            ["stop_cmd"]      = "/sbin/service httpd stop",
            ["reload_cmd"]    = "/sbin/service httpd reload",
            ["restart_cmd"]   = "/sbin/service httpd graceful",
            ["available_dir"] = nil,
            ["enabled_dir"]   = "/etc/httpd/vhosts.d",
            ["enable_cmd"]    = nil,
            ["disable_cmd"]   = nil,
            ["httpd_user"]    = "apache",
            ["httpd_group"]   = "apache",
            ["ssl_chain_dir"] = "/etc/pki/tls/certs",
            ["ssl_cert_dir"]  = "/etc/pki/tls/certs",
            ["ssl_key_dir"]   = "/etc/pki/tls/private",
            ["ssl_key_user"]  = "root",
            ["ssl_key_group"] = "root",
            ["ssl_key_mode"]  = "0600",
            ["htdocs_path"]   = LC.web.HTDOCS_PATH,
            ["modules"]       = check_modules("/usr/sbin/httpd", "-M"),
            ["run_dir"]       = "/var/run/httpd",
            ["reload_max"]    = RELOAD_MAX,
            ["units"]         = { ["apache"] = "httpd" }
        }
        check_http2(data)
        return data
      end
    -- else: fall trough, to check for custom Apache installation
      LC.log.print(LC.log.ERROR, "LC.apache.detect(): Found Apache package '", pkg, "', but no binary at ", bin)
    end
  elseif LC.distribution.family == "SUSE" then
    -- OpenSUSE
    local pkg, v = LC.distribution.hasPackage('apache2')
    if pkg ~= nil then
      LC.log.print(LC.log.DEBUG, "Found package '", pkg, "' (Version ", v, ")")
      local bin = "/usr/sbin/httpd2"
      if not LC.fs.is_file(bin) then
        bin = "/usr/sbin/httpd"
      end
      local bv = getBinaryVersion(bin)
      LC.log.print(LC.log.DEBUG, "binver ", bv)
      if bv ~= nil then
        local data = {
            ["type"]          = "apache",
            ["binname"]       = bin,
            ["binversion"]    = bv,
            ["pkgname"]       = pkg,
            ["pkgversion"]    = v,
            ["configpath"]    = "/etc/apache2",
            ["configfile"]    = "/etc/apache2/conf.d/liveconfig.conf",
            ["statusfile"]    = "/etc/apache2/liveconfig.status",
            ["start_cmd"]     = "/usr/sbin/rcapache2 start",
            ["stop_cmd"]      = "/usr/sbin/rcapache2 stop",
            ["reload_cmd"]    = "/usr/sbin/rcapache2 reload",
            ["restart_cmd"]   = "/usr/sbin/rcapache2 restart",
            ["available_dir"] = nil,
            ["enabled_dir"]   = "/etc/apache2/vhosts.d",
            ["enable_cmd"]    = nil,
            ["disable_cmd"]   = nil,
            ["httpd_user"]    = "wwwrun",
            ["httpd_group"]   = "www",
            ["ssl_chain_dir"] = "/etc/apache2/ssl.crt",
            ["ssl_cert_dir"]  = "/etc/apache2/ssl.crt",
            ["ssl_key_dir"]   = "/etc/apache2/ssl.key",
            ["ssl_key_user"]  = "root",
            ["ssl_key_group"] = "root",
            ["ssl_key_mode"]  = "0600",
            ["htdocs_path"]   = LC.web.HTDOCS_PATH,
            ["modules"]       = check_modules("/usr/sbin/apache2ctl", "-M"),
            ["run_dir"]       = "/var/run",
            ["reload_max"]    = RELOAD_MAX,
            ["units"]         = { ["apache"] = "apache2" }
        }
        check_http2(data)
        return data
      end
    -- else: fall trough, to check for custom Apache installation
      LC.log.print(LC.log.ERROR, "LC.apache.detect(): Found Apache package '", pkg, "', but no binary at ", bin)
    end
  end  
end

-- ---------------------------------------------------------------------------
-- install(cfg, opts)
--
-- Install LiveConfig configuration files for Apache httpd
-- ---------------------------------------------------------------------------
function install(cfg, opts)
  LC.log.print(LC.log.INFO, "Activating management of Apache httpd")

  local fh, status, msg
  local configpath = cfg["configpath"]
  local configfile = cfg["configfile"]
  local statusfile = cfg["statusfile"]
  local vhostpath  = cfg["available_dir"] or cfg["enabled_dir"]
  local vhostfile  -- set below (distribution-specific)
  local v_major, v_minor, v_patch = string.match(cfg["binversion"], "^(%d+)%.(%d+)%.(%d+)")
  v_major = tonumber(v_major)
  v_minor = tonumber(v_minor)
  v_patch = tonumber(v_patch)

  if opts and opts.prefix then
    -- use prefix (for testing etc.)
    configpath = opts.prefix .. configpath
    configfile = opts.prefix .. configfile
    statusfile = opts.prefix .. statusfile
    vhostpath  = opts.prefix .. vhostpath
  end

  if LC.distribution.family == "RedHat" then

    -- create httpd.conf for RedHat
    local apachecfgfile = configpath .. "/conf/httpd.conf"

    -- backup old apachecfgfile
    if LC.fs.is_file(apachecfgfile) and not LC.fs.is_file(apachecfgfile .. ".lcbak") then
      LC.fs.rename(apachecfgfile, apachecfgfile .. ".lcbak")
    end

    -- backup old ssl.conf
    if LC.fs.is_file(configpath .. "/conf.d/ssl.conf") then
      LC.fs.rename(configpath .. "/conf.d/ssl.conf", configpath .. "/conf.d/ssl.conf.lcbak")
    end

    fh, msg = io.open(apachecfgfile, "w")
    if fh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", apachecfgfile, "' for writing: ", msg)
      return false, "Can't open '" .. apachecfgfile .. "' for writing: " .. msg
    end

    -- write header
    LC.liveconfig.writeHeader(fh)

    fh:write( [[

ServerTokens Prod
ServerRoot "/etc/httpd"
PidFile run/httpd.pid
Timeout 60
KeepAlive On
MaxKeepAliveRequests 100
KeepAliveTimeout 15
]])

    if LC.fs.is_dir("/etc/httpd/conf.modules.d") then
      fh:write("Include conf.modules.d/*.conf\n")
    else
      fh:write( [[
<IfModule prefork.c>
StartServers       8
MinSpareServers    5
MaxSpareServers   20
ServerLimit      256
MaxClients       256
MaxRequestsPerChild  4000
</IfModule>

<IfModule worker.c>
StartServers         2
MaxClients         150
MinSpareThreads     25
MaxSpareThreads     75 
ThreadsPerChild     25
MaxRequestsPerChild  0
</IfModule>

LoadModule auth_basic_module modules/mod_auth_basic.so
LoadModule auth_digest_module modules/mod_auth_digest.so
LoadModule authn_file_module modules/mod_authn_file.so
LoadModule authn_alias_module modules/mod_authn_alias.so
LoadModule authn_anon_module modules/mod_authn_anon.so
LoadModule authn_dbm_module modules/mod_authn_dbm.so
LoadModule authn_default_module modules/mod_authn_default.so
LoadModule authz_host_module modules/mod_authz_host.so
LoadModule authz_user_module modules/mod_authz_user.so
LoadModule authz_owner_module modules/mod_authz_owner.so
LoadModule authz_groupfile_module modules/mod_authz_groupfile.so
LoadModule authz_dbm_module modules/mod_authz_dbm.so
LoadModule authz_default_module modules/mod_authz_default.so
LoadModule ldap_module modules/mod_ldap.so
LoadModule authnz_ldap_module modules/mod_authnz_ldap.so
LoadModule include_module modules/mod_include.so
LoadModule log_config_module modules/mod_log_config.so
LoadModule logio_module modules/mod_logio.so
LoadModule env_module modules/mod_env.so
LoadModule ext_filter_module modules/mod_ext_filter.so
LoadModule mime_magic_module modules/mod_mime_magic.so
LoadModule expires_module modules/mod_expires.so
LoadModule deflate_module modules/mod_deflate.so
LoadModule headers_module modules/mod_headers.so
LoadModule usertrack_module modules/mod_usertrack.so
LoadModule setenvif_module modules/mod_setenvif.so
LoadModule mime_module modules/mod_mime.so
LoadModule dav_module modules/mod_dav.so
LoadModule status_module modules/mod_status.so
LoadModule autoindex_module modules/mod_autoindex.so
LoadModule info_module modules/mod_info.so
LoadModule dav_fs_module modules/mod_dav_fs.so
LoadModule vhost_alias_module modules/mod_vhost_alias.so
LoadModule negotiation_module modules/mod_negotiation.so
LoadModule dir_module modules/mod_dir.so
LoadModule actions_module modules/mod_actions.so
LoadModule speling_module modules/mod_speling.so
LoadModule userdir_module modules/mod_userdir.so
LoadModule alias_module modules/mod_alias.so
LoadModule rewrite_module modules/mod_rewrite.so
LoadModule proxy_module modules/mod_proxy.so
LoadModule proxy_balancer_module modules/mod_proxy_balancer.so
LoadModule proxy_ftp_module modules/mod_proxy_ftp.so
LoadModule proxy_http_module modules/mod_proxy_http.so
LoadModule proxy_connect_module modules/mod_proxy_connect.so
LoadModule cache_module modules/mod_cache.so
LoadModule suexec_module modules/mod_suexec.so
LoadModule disk_cache_module modules/mod_disk_cache.so
# LoadModule file_cache_module modules/mod_file_cache.so
# LoadModule mem_cache_module modules/mod_mem_cache.so
LoadModule cgi_module modules/mod_cgi.so
LoadModule version_module modules/mod_version.so
]] )
    end

    fh:write( [[

Include conf.d/*.conf

User apache
Group apache
ServerAdmin root@localhost
UseCanonicalName Off
DocumentRoot "/var/www/html"
<Directory />
    Options FollowSymLinks
    AllowOverride None
</Directory>

<Directory "/var/www/html">
    Options Indexes FollowSymLinks
    AllowOverride None
    Order allow,deny
    Allow from all
</Directory>

<IfModule mod_userdir.c>
    #
    # UserDir is disabled by default since it can confirm the presence
    # of a username on the system (depending on home directory
    # permissions).
    #
    UserDir disable

    #
    # To enable requests to /~user/ to serve the user's public_html
    # directory, remove the "UserDir disable" line above, and uncomment
    # the following line instead:
    # 
    #UserDir public_html

</IfModule>

DirectoryIndex index.html index.html.var
AccessFileName .htaccess

<Files ~ "^\.ht">
    Order allow,deny
    Deny from all
</Files>

TypesConfig /etc/mime.types
DefaultType text/plain

<IfModule mod_mime_magic.c>

    MIMEMagicFile conf/magic
</IfModule>

HostnameLookups Off
ErrorLog logs/error_log
LogLevel warn
ServerSignature On

]])

    if not LC.fs.is_file("/etc/httpd/conf.d/autoindex.conf") then
      fh:write( [[
Alias /icons/ "/var/www/icons/"

<Directory "/var/www/icons">
    Options Indexes MultiViews
    AllowOverride None
    Order allow,deny
    Allow from all
</Directory>

IndexOptions FancyIndexing VersionSort NameWidth=* HTMLTable
AddIconByEncoding (CMP,/icons/compressed.gif) x-compress x-gzip

AddIconByType (TXT,/icons/text.gif) text/*
AddIconByType (IMG,/icons/image2.gif) image/*
AddIconByType (SND,/icons/sound2.gif) audio/*
AddIconByType (VID,/icons/movie.gif) video/*

AddIcon /icons/binary.gif .bin .exe
AddIcon /icons/binhex.gif .hqx
AddIcon /icons/tar.gif .tar
AddIcon /icons/world2.gif .wrl .wrl.gz .vrml .vrm .iv
AddIcon /icons/compressed.gif .Z .z .tgz .gz .zip
AddIcon /icons/a.gif .ps .ai .eps
AddIcon /icons/layout.gif .html .shtml .htm .pdf
AddIcon /icons/text.gif .txt
AddIcon /icons/c.gif .c
AddIcon /icons/p.gif .pl .py
AddIcon /icons/f.gif .for
AddIcon /icons/dvi.gif .dvi
AddIcon /icons/uuencoded.gif .uu
AddIcon /icons/script.gif .conf .sh .shar .csh .ksh .tcl
AddIcon /icons/tex.gif .tex
AddIcon /icons/bomb.gif core

AddIcon /icons/back.gif ..
AddIcon /icons/hand.right.gif README
AddIcon /icons/folder.gif ^^DIRECTORY^^
AddIcon /icons/blank.gif ^^BLANKICON^^
DefaultIcon /icons/unknown.gif
ReadmeName README.html
HeaderName HEADER.html
IndexIgnore .??* *~ *# HEADER* README* RCS CVS *,v *,t

]])
    end

    fh:write( [[

<IfModule mod_dav_fs.c>
    # Location of the WebDAV lock database.
    DAVLockDB /var/lib/dav/lockdb
</IfModule>

ScriptAlias /cgi-bin/ "/var/www/cgi-bin/"
<Directory "/var/www/cgi-bin">
    AllowOverride None
    Options None
    Order allow,deny
    Allow from all
</Directory>
AddLanguage ca .ca
AddLanguage cs .cz .cs
AddLanguage da .dk
AddLanguage de .de
AddLanguage el .el
AddLanguage en .en
AddLanguage eo .eo
AddLanguage es .es
AddLanguage et .et
AddLanguage fr .fr
AddLanguage he .he
AddLanguage hr .hr
AddLanguage it .it
AddLanguage ja .ja
AddLanguage ko .ko
AddLanguage ltz .ltz
AddLanguage nl .nl
AddLanguage nn .nn
AddLanguage no .no
AddLanguage pl .po
AddLanguage pt .pt
AddLanguage pt-BR .pt-br
AddLanguage ru .ru
AddLanguage sv .sv
AddLanguage zh-CN .zh-cn
AddLanguage zh-TW .zh-tw

LanguagePriority en ca cs da de el eo es et fr he hr it ja ko ltz nl nn no pl pt pt-BR ru sv zh-CN zh-TW
ForceLanguagePriority Prefer Fallback
AddDefaultCharset UTF-8
AddType application/x-compress .Z
AddType application/x-gzip .gz .tgz
AddHandler type-map var
AddType text/html .shtml
AddOutputFilter INCLUDES .shtml

Alias /error/ "/var/www/error/"

<IfModule mod_negotiation.c>
<IfModule mod_include.c>
    <Directory "/var/www/error">
        AllowOverride None
        Options IncludesNoExec
        AddOutputFilter Includes html
        AddHandler type-map var
        Order allow,deny
        Allow from all
        LanguagePriority en es de fr
        ForceLanguagePriority Prefer Fallback
    </Directory>

</IfModule>
</IfModule>

BrowserMatch "Mozilla/2" nokeepalive
BrowserMatch "MSIE 4\.0b2;" nokeepalive downgrade-1.0 force-response-1.0
BrowserMatch "RealPlayer 4\.0" force-response-1.0
BrowserMatch "Java/1\.0" force-response-1.0
BrowserMatch "JDK/1\.0" force-response-1.0
BrowserMatch "Microsoft Data Access Internet Publishing Provider" redirect-carefully
BrowserMatch "MS FrontPage" redirect-carefully
BrowserMatch "^WebDrive" redirect-carefully
BrowserMatch "^WebDAVFS/1.[0123]" redirect-carefully
BrowserMatch "^gnome-vfs/1.0" redirect-carefully
BrowserMatch "^XML Spy" redirect-carefully
BrowserMatch "^Dreamweaver-WebDAV-SCM1" redirect-carefully

# Include vhost configurations
Include /etc/httpd/vhosts.d/*.conf
]] )
    -- write footer
    LC.liveconfig.writeFooter(fh)
    fh:close()

    -- create new ssl.conf file:
    fh, msg = io.open(configpath .. "/conf.d/ssl.conf", "w")
    if fh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", configpath, "/conf.d/ssl.conf' for writing: ", msg)
      return false, "Can't open '" .. configpath .. "/conf.d/ssl.conf' for writing: " .. msg
    end

    -- write header
    LC.liveconfig.writeHeader(fh)

    fh:write( [[

LoadModule ssl_module modules/mod_ssl.so
SSLPassPhraseDialog  builtin
SSLSessionCache         shmcb:/var/cache/mod_ssl/scache(512000)
SSLSessionCacheTimeout  300
SSLRandomSeed startup file:/dev/urandom  256
SSLRandomSeed connect builtin
SSLCryptoDevice builtin
]] )

    if v_major > 2 or (v_major == 2 and v_minor > 3) or (v_major == 2 and v_minor == 3 and v_patch > 3) then
      fh:write("Mutex default\n")
    else
      fh:write("SSLMutex default\n")
    end

    -- write footer
    LC.liveconfig.writeFooter(fh)
    fh:close()

  end -- if LC.distribution.family == "RedHat"

  -- create directory for vhosts:
  if not LC.fs.is_dir(vhostpath) then
    LC.fs.mkdir(vhostpath)
    LC.fs.setperm(vhostpath, 750, "root", cfg.httpd_group)
  end

  -- create 'liveconfig.conf' config file
  fh, msg = io.open(configfile, "w")
  if fh == nil then
    LC.log.print(LC.log.ERR, "Can't open '", configfile, "' for writing: ", msg)
    return false, "Can't open '" .. configfile .. "' for writing: " .. msg
  end

  -- write header
  LC.liveconfig.writeHeader(fh)

  -- write config...
  fh:write([[

# Access logging
# We define a special log format here, which gets piped into the 'lclogsplit'
# utility (see lclogsplit(1) for more details).
# This way Apache httpd needs only one file handle for all access_log files,
# and we get nice real-time statistics on HTTP traffic.
# IMPORTANT: the module 'mod_logio' needs to be enabled!

<IfModule mod_logio.c>
  LogFormat "%v:#:%a %l %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-agent}i\" %I %O" LiveConfig
</IfModule>

<IfModule !mod_logio.c>
  LogFormat "%v:#:%a %l %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-agent}i\"" LiveConfig
</IfModule>

]])
  if v_major > 2 or (v_major == 2 and v_minor > 2) or (v_major == 2 and v_minor == 2 and v_patch >= 11) then
    fh:write("CustomLog \"||")
  else
    -- Apache supports '||' prefix for CustomLog only from 2.2.11+; earlier versions need "|exec":
    fh:write("CustomLog \"|exec ")
  end
  fh:write(LC.liveconfig.libexecdir, "/lclogsplit -i -w\" LiveConfig\n\n")

  if LC.distribution.family == "Debian" then
    -- for Debian, add global "ErrorLog"
    fh:write("ErrorLog /var/log/apache2/error.log\n\n")
  end
  -- RedHat/CentOS has its ErrorLog directove in /etc/httpd/httpd.conf
  -- Gentoo has its ErrorLog directive in /etc/apache2/modules.d/00_default_settings.conf

  -- write footer
  LC.liveconfig.writeFooter(fh)

  fh:close()

  if LC.fs.is_dir("/etc/apache2/conf-enabled") and LC.fs.is_file("/usr/sbin/a2enconf") then
    -- enable configuration
    LC.exec("/usr/sbin/a2enconf liveconfig")
  end

  if LC.distribution.family == "Debian" then
    if LC.fs.is_file(configpath .. "/ports.conf") then
      -- Debian has a "ports.conf" file, containing the "Listen" directives
      -- for Apache. Remove this, and add a note that Listen commands are now
      -- in the default vhost config file
      if not LC.fs.is_file(configpath .. "/ports.conf.lcbak") then
        -- only back up if not done yet
        LC.fs.rename(configpath .. "/ports.conf", configpath .. "/ports.conf.lcbak")
      end
      fh, msg = io.open(configpath .. "/ports.conf", "w")
      if fh == nil then
        LC.log.print(LC.log.ERR, "Can't open '", configpath .. "/ports.conf", "' for writing: ", msg)
        return false, "Can't open '" .. configpath .. "/ports.conf", "' for writing: " .. msg
      end
      LC.liveconfig.writeHeader(fh)
      fh:write([[

# The "Listen" and "NameVirtualHost" directives are now in the first
# vhost configuration file (usually /etc/apache2/sites-available/default)
#
# The original "ports.conf" file has been backed up at "ports.conf.lcbak"

]])
      LC.liveconfig.writeFooter(fh)
      fh:close()
    end
    if LC.fs.is_file(vhostpath .. "/000-default.conf") then
      vhostfile = "000-default.conf"
    else
      vhostfile = "default"
    end
  elseif LC.distribution.family == "Gentoo" then
    vhostfile = "00_default_vhost.conf"
    -- remove default ssl vhost file (if existing)
    if LC.fs.is_file(vhostpath .. "/00_default_ssl_vhost.conf") then
      LC.fs.rename(vhostpath .. "/00_default_ssl_vhost.conf", vhostpath .. "/00_default_ssl_vhost.conf.lcbak")
    end
  elseif LC.distribution.family == "RedHat" then
    vhostfile = "00_default_vhost.conf"
  elseif LC.distribution.family == "SUSE" then
    if LC.fs.is_file(configpath .. "/listen.conf") then
      -- SUSE has a "listen.conf" file, containing the "Listen" directives
      -- for Apache. Remove this, and add a note that Listen commands are now
      -- in the default vhost config file
      if not LC.fs.is_file(configpath .. "/listen.conf.lcbak") then
        -- only back up if not done yet
        LC.fs.rename(configpath .. "/listen.conf", configpath .. "/listen.conf.lcbak")
      end
      fh, msg = io.open(configpath .. "/listen.conf", "w")
      if fh == nil then
        LC.log.print(LC.log.ERR, "Can't open '", configpath .. "/listen.conf", "' for writing: ", msg)
        return false, "Can't open '" .. configpath .. "/listen.conf", "' for writing: " .. msg
      end
      LC.liveconfig.writeHeader(fh)
      fh:write([[

# The "Listen" and "NameVirtualHost" directives are now in the first
# vhost configuration file (usually /etc/apache2/vhosts.d/default.conf)
#
# The original "listen.conf" file has been backed up at "listen.conf.lcbak"

]])
      LC.liveconfig.writeFooter(fh)
      fh:close()
    end
    vhostfile = "default.conf"
  else
    LC.log.print(LC.log.ERR, "Distribution family '", LC.distribution.family, "' not supported (yet)")
    return false, "Distribution family '" .. LC.distribution.family .. "' not supported (yet)"
  end

  -- back up old default vhost file (if existing)
  if LC.fs.is_file(vhostpath .. "/" .. vhostfile) and not LC.fs.is_file(vhostpath .. "/" .. vhostfile .. ".lcbak") then
    LC.fs.rename(vhostpath .. "/" .. vhostfile, vhostpath .. "/" .. vhostfile .. ".lcbak")
  end

  -- write default vhost file:
  configure(cfg, opts)

  -- set permissions for lclogsplit under RedHat
  if LC.distribution.family == "RedHat" then
    LC.exec("chcon -v --type=httpd_exec_t " .. LC.liveconfig.libexecdir .. "/lclogsplit")
  end
  -- write status file
  LC.liveconfig.writeStatus(statusfile, cfg.type, 0, os.date("%Y-%m-%d %H:%M:%S %Z"))

  return true
end

-- ---------------------------------------------------------------------------
-- uninstall(cfg, opts)
--
-- Remove LiveConfig configuration files from Apache httpd
-- ---------------------------------------------------------------------------
function uninstall(cfg, opts)
  LC.log.print(LC.log.INFO, "Disabling management of Apache httpd")

  local status, msg
  local configpath = cfg["configpath"]
  local configfile = cfg["configfile"]
  local statusfile = cfg["statusfile"]
  local vhostpath  = cfg["available_dir"] or cfg["enabled_dir"]

  if opts and opts.prefix then
    -- use prefix (for testing etc.)
    configpath = opts.prefix .. configpath
    configfile = opts.prefix .. configfile
    statusfile = opts.prefix .. statusfile
    vhostpath = opts.prefix .. vhostpath
  end

  -- remove config file
  status, msg = os.remove(configfile)
  if status == nil then
    LC.log.print(LC.log.ERR, "Deletion of '", configfile, "' failed: ", msg)
  end

  if LC.fs.is_dir("/etc/apache2/conf-enabled") and LC.fs.is_file("/usr/sbin/a2disconf") then
    -- disable configuration
    LC.exec("/usr/sbin/a2disconf liveconfig")
  end

  if LC.distribution.family == "Debian" then
    if LC.fs.is_file(vhostpath .. "/000-default.conf") then
      vhostfile = "000-default.conf"
    else
      vhostfile = "default"
    end
    -- restore original "ports.conf" (if existing...)
    if LC.fs.is_file(configpath .. "/ports.conf") and LC.fs.is_file(configpath .. "/ports.conf.lcbak") then
      LC.fs.rename(configpath .. "/ports.conf.lcbak", configpath .. "/ports.conf")
    end
  elseif LC.distribution.family == "RedHat" then
    vhostfile = "00_default_vhost.conf"
    -- restore original "httpd.conf" (if existing...)
    if LC.fs.is_file(configpath .. "/httpd.conf") and LC.fs.is_file(configpath .. "/httpd.conf.lcbak") then
      LC.fs.rename(configpath .. "/httpd.conf.lcbak", configpath .. "/httpd.conf")
    end
    -- restore original "ssl.conf" (if existing...)
    if LC.fs.is_file(configpath .. "/conf.d/ssl.conf.lcbak") then
      LC.fs.rename(configpath .. "/conf.d/ssl.conf.lcbak", configpath .. "/conf.d/ssl.conf")
    end
  elseif LC.distribution.family == "SUSE" then
    vhostfile = "default.conf"
    -- restore original "listen.conf" (if existing...)
    if LC.fs.is_file(configpath .. "/listen.conf") and LC.fs.is_file(configpath .. "/listen.conf.lcbak") then
      LC.fs.rename(configpath .. "/listen.conf.lcbak", configpath .. "/listen.conf")
    end
    -- remove our "default.conf" file
    if LC.fs.is_file(vhostpath .. "/" .. vhostfile) then
      os.remove(vhostpath .. "/" .. vhostfile)
    end
  elseif LC.distribution.family == "Gentoo" then
    vhostfile = "00_default_vhost.conf"
  else
    LC.log.print(LC.log.ERR, "Distribution family '", LC.distribution.family, "' not supported (yet)")
    return false, "Distribution family '" .. LC.distribution.family .. "' not supported (yet)"
  end

  -- restore old default vhost file (if existing)
  if LC.fs.is_file(vhostpath .. "/" .. vhostfile .. ".lcbak") then
    LC.fs.rename(vhostpath .. "/" .. vhostfile .. ".lcbak", vhostpath .. "/" .. vhostfile)
  end

  -- restore default ssl vhost file
  if LC.distribution.family == "Gentoo" then
      if LC.fs.is_file(vhostpath .. "/00_default_ssl_vhost.conf.lcbak") then
        LC.fs.rename(vhostpath .. "/00_default_ssl_vhost.conf.lcbak", vhostpath .. "/00_default_ssl_vhost.conf")
      end
  end

  -- remove status file
  if LC.fs.is_file(statusfile) then
      os.remove(statusfile)
  end

  return true
end

-- ---------------------------------------------------------------------------
local function updateCRTfile(cfg, crtfilename, crt)
  local sfh, msg
  sfh, msg = io.open(crtfilename .. ".tmp", "w")
  if sfh == nil then
    LC.log.print(LC.log.ERR, "Can't open '", crtfilename, ".tmp' for writing: ", msg)
    return false, "Can't open '" .. crtfilename .. ".tmp' for writing: " .. msg
  end
  sfh:write(crt)
  -- add linebreak (if supplied certificate doesn't have one)
  sfh:write("\n")

  -- custom DH parameters (Logjam attack)
  if not LC.fs.is_file(cfg.configpath .. "/dhparam.pem") and LC.fs.is_file("/etc/liveconfig/dhparam.dist.pem") then
    -- copy non-standard default DH parameters until local parameters have been generated:
    LC.fs.copy("/etc/liveconfig/dhparam.dist.pem", cfg.configpath .. "/dhparam.pem")
    LC.fs.setperm(cfg.configpath .. "/dhparam.pem", "0640", "root", "root")
  end

  -- append custom DH parameters
  if LC.fs.is_file(cfg.configpath .. "/dhparam.pem") then
    local fhr = io.open(cfg.configpath .. "/dhparam.pem", "r")
    if fhr ~= nil then
      local line
      sfh:write("# custom DH parameters (copied from ", cfg.configpath, "/dhparam.pem)\n")
      while true do
        line = fhr:read()
        if line == nil then break end
        sfh:write(line, "\n")
      end
      fhr:close()
    end
  end

  sfh:close()
  LC.fs.setperm(crtfilename .. ".tmp", "0644", "root", "root")
  LC.fs.rename(crtfilename .. ".tmp", crtfilename)
  return crtfilename
end

-- ---------------------------------------------------------------------------
-- configure(cfg, opts)
--
-- Configure Apache httpd
-- ---------------------------------------------------------------------------
function configure(cfg, opts)

  local fh, status, msg
  local configpath = cfg["configpath"]
  local statusfile = cfg["statusfile"]
  local vhostpath  = cfg["available_dir"] or cfg["enabled_dir"]
  local vhostfile  -- set below (distribution-specific)
  local v_major, v_minor, v_patch = string.match(cfg["binversion"], "^(%d+)%.(%d+)%.(%d+)")
  v_major = tonumber(v_major)
  v_minor = tonumber(v_minor)
  v_patch = tonumber(v_patch)

  if opts and opts.prefix then
    -- use prefix (for testing etc.)
    configpath = opts.prefix .. configpath
    statusfile = opts.prefix .. statusfile
    vhostpath  = opts.prefix .. vhostpath
  end

  if LC.distribution.family == "Debian" then
    if LC.fs.is_file(vhostpath .. "/000-default.conf") or LC.fs.is_file(vhostpath .. "/000-default.conf.lcbak") then
      vhostfile = "000-default.conf"
    else
      vhostfile = "default"
    end
  elseif LC.distribution.family == "Gentoo" or LC.distribution.family == "RedHat" then
    vhostfile = "00_default_vhost.conf"
  elseif LC.distribution.family == "SUSE" then
    vhostfile = "default.conf"
  else
    LC.log.print(LC.log.ERR, "Distribution family '", LC.distribution.family, "' not supported (yet)")
    return false, "Distribution family '" .. LC.distribution.family .. "' not supported (yet)"
  end

  fh, msg = io.open(vhostpath .. "/" .. vhostfile .. ".tmp", "w")
  if fh == nil then
    LC.log.print(LC.log.ERR, "Can't open '", vhostpath .. "/" .. vhostfile, "' for writing: ", msg)
    return false, "Can't open '" .. vhostpath .. "/" .. vhostfile.. "' for writing: " .. msg
  end
  LC.liveconfig.writeHeader(fh)

  fh:write([[
# Default VirtualHost configuration
# Automatically created by LiveConfig

]])

  local interfaces = "", i, s
  local CONF = { }

  -- SSL configuration
  if opts.ssl_interfaces ~= nil then
    fh:write("<IfModule mod_ssl.c>\n")
    CONF["SSLHonorCipherOrder"] = "on"
    local protocols
    if opts.ssl_strong then
      protocols = SSL_PROTOCOLS_STRONG
      CONF["SSLCipherSuite"] = LC.liveconfig.DEFAULT_SSL_STRONG_CIPHERS
    else
      protocols = SSL_PROTOCOLS
      CONF["SSLCipherSuite"] = LC.liveconfig.DEFAULT_SSL_CIPHERS
    end

    if opts.ssl_notls1 then
      -- disable TLS 1.0 and 1.1
      if not protocols:match('-TLSv1 ') and not protocols:match('-TLSv1$') then
        protocols = protocols .. ' -TLSv1'
      end
      if not protocols:match('-TLSv1%.1 ') and not protocols:match('-TLSv1%.1$') then
        protocols = protocols .. ' -TLSv1.1'
      end
    end
    CONF["SSLProtocol"] = protocols

    if v_major > 2 or (v_major == 2 and (v_minor > 3 or (v_minor == 3 and v_patch >= 3))) then
      -- configure OCSP Stapling with Apache >=2.3.3
      CONF["SSLStaplingReturnResponderErrors"] = "off"
      CONF["SSLStaplingResponderTimeout"] = "5"
      CONF["SSLStaplingCache"] = "shmcb:" .. cfg.run_dir .. "/ssl_stapling_cache(128000)"
    end
    for i=1, #opts.ssl_interfaces do
      fh:write("  Listen ", opts.ssl_interfaces[i], "\n")
      if v_major < 2 or (v_major == 2 and v_minor < 4) then
        fh:write("  NameVirtualHost ", opts.ssl_interfaces[i], "\n")
      end
    end

    -- custom DH parameters (Logjam attack)
    if not LC.fs.is_file(configpath .. "/dhparam.pem") and LC.fs.is_file("/etc/liveconfig/dhparam.dist.pem") then
      -- copy non-standard default DH parameters until local parameters have been generated:
      LC.fs.copy("/etc/liveconfig/dhparam.dist.pem", configpath .. "/dhparam.pem")
      LC.fs.setperm(configpath .. "/dhparam.pem", "0640", "root", "root")
    end
    --[[
    -- DISABLED - only works with Apache 2.4.8+ *AND* OpenSSL 1.0.2+, which is NOT available on Debian 8!
    -- SSLOpenSSLConfCmd requires httpd v2.4.8+ (http://httpd.apache.org/docs/trunk/mod/mod_ssl.html#sslopensslconfcmd)
    if LC.fs.is_file(configpath .. "/dhparam.pem") and (v_major > 2 or (v_major == 2 and v_minor > 4) or (v_major == 2 and v_minor == 4 and v_patch >= 8)) then
      if LC.fs.is_file(configpath .. "/dhparam.pem") then
        fh:write("  SSLOpenSSLConfCmd DHParam \"", configpath, "/dhparam.pem\"\n")
      end
    end
    ]]--
    -- WORKAROUND: append custom DH parameters to every SSLCertificateFile, this is backwards-compatible

    -- merge with local (override) configuration settings
    for i,s in pairs(SSLDEFAULT) do CONF[i]=s end

    -- sort configuration table (for human readers)
    local confkeys = {}
    for i in pairs(CONF) do table.insert(confkeys, i) end
    table.sort(confkeys)

    -- write configuration options:
    for i, s in ipairs(confkeys) do
      fh:write("  ", s, " ", CONF[s], "\n")
    end

    fh:write("</IfModule>\n\n")
  end -- HTTPS interfaces

  local is_revproxy = false     -- set to "true" if Apache is configured as reverse proxy target

  if opts.interfaces == nil then
    fh:write("# NO HTTP INTERFACES CONFIGURED!\n")
  else
    for i=1, #opts.interfaces do
      interfaces = interfaces .. " " .. opts.interfaces[i]
      fh:write("Listen ", opts.interfaces[i], "\n")
      if v_major < 2 or (v_major == 2 and v_minor < 4) then
        fh:write("NameVirtualHost ", opts.interfaces[i], "\n")
      end
      if opts.interfaces[i] == "127.0.0.1:80" then
        is_revproxy = true
      end
    end
  end

  if opts.interfaces ~= nil or opts.ssl_interfaces ~= nil then
    if is_revproxy then
      fh:write([[

<IfModule mod_remoteip.c>
  RemoteIPHeader X-Forwarded-For
  RemoteIPInternalProxy 127.0.0.1
</IfModule>
<IfModule mod_rpaf.c>
  RPAFenable On
  RPAFsethostname Off
  RPAFheader X-Forwarded-For
  RPAFproxy_ips 127.0.0.1
</IfModule>
<IfModule mod_setenvif.c>
  SetEnvIf X-Forwarded-Proto https HTTPS=on
</IfModule>
]])
    end
    fh:write([[

ServerName default
DocumentRoot "]], LC.liveconfig.datadir, [[/html/"

Alias /.errorFiles/ ]], LC.liveconfig.datadir, [[/html/
<Directory "]], LC.liveconfig.datadir, [[/html/">
    Options SymLinksIfOwnerMatch IncludesNOEXEC
    AllowOverride None
]])
    if v_major < 2 or (v_major == 2 and v_minor < 3) then
      fh:write("    Order allow,deny\n")
      fh:write("    allow from all\n")
    else
      fh:write("    <IfModule mod_access_compat.c>\n")
      fh:write("        Order allow,deny\n")
      fh:write("        Allow from all\n")
      fh:write("    </IfModule>\n")
      fh:write("    <IfModule mod_authz_core.c>\n")
      fh:write("        Require all granted\n")
      fh:write("    </IfModule>\n")
    end
    fh:write([[
    <IfModule mod_rewrite.c>
        RewriteEngine On
        RewriteRule ^$ - [R=404,L]
    </IfModule>
    <IfModule mod_include.c>
        AddType text/html .shtml
        AddOutputFilter INCLUDES .shtml
        ErrorDocument 404 /not-available.shtml
        <IfModule !mod_rewrite.c>
            DirectoryIndex not-available.shtml
        </IfModule>
    </IfModule>
    <IfModule !mod_include.c>
        ErrorDocument 404 /not-available.html
        <IfModule !mod_rewrite.c>
            DirectoryIndex not-available.html
        </IfModule>
    </IfModule>
</Directory>

<Directory ]], LC.web.getWebRoot(), [[/.well-known/htdocs/>
    Options Indexes FollowSymLinks
    AllowOverride None
]])
    if v_major < 2 or (v_major == 2 and v_minor < 3) then
      fh:write("    Order allow,deny\n")
      fh:write("    allow from all\n")
    else
      fh:write("    <IfModule mod_access_compat.c>\n")
      fh:write("        Order allow,deny\n")
      fh:write("        Allow from all\n")
      fh:write("    </IfModule>\n")
      fh:write("    <IfModule mod_authz_core.c>\n")
      fh:write("        Require all granted\n")
      fh:write("    </IfModule>\n")
    end
    fh:write([[
</Directory>

]])

  end

  if opts.interfaces ~= nil then
    -- default HTTP host
    fh:write([[

<VirtualHost]], interfaces, [[>

    Servername default
    DocumentRoot "]], LC.liveconfig.datadir, [[/html/"

    <IfModule mod_rewrite.c>
        RewriteEngine On
        RewriteMap  "lc" "int:tolower"
        RewriteCond %{REQUEST_URI} ^/\.well-known/(.*/)?([^/]*)$
        RewriteCond ]], LC.liveconfig.datadir, [[/html/.well-known/%1%2 !-f
        RewriteCond ]], LC.web.getWebRoot(), [[/.well-known/htdocs/%1${lc:%{HTTP_HOST}}.%2 -f
        RewriteRule / ]], LC.web.getWebRoot(), [[/\.well-known/htdocs/%1${lc:%{HTTP_HOST}}.%2 [T=text/plain,L]
    </IfModule>

]])

    if LC.fs.is_file(vhostpath .. "/default.inc") then
      fh:write("    Include ", vhostpath, "/default.inc\n")
    end

    fh:write("</VirtualHost>\n\n")

  end

  if opts.ssl_sni ~= nil and opts.ssl_sni.interfaces ~= nil then
    -- configure shared SSL interfaces (SNI) to use default vHost with default SSL certificate
    local sni_interfaces = ""
    for i=1, #opts.ssl_sni.interfaces do
      sni_interfaces = sni_interfaces .. " " .. opts.ssl_sni.interfaces[i]
    end
    fh:write("<IfModule mod_ssl.c>\n")
    fh:write("<VirtualHost", sni_interfaces, ">\n\n")
    fh:write("    ServerName default\n")
    fh:write("    DocumentRoot \"", LC.liveconfig.datadir, "/html/\"\n\n")
    fh:write("    SSLEngine On\n")

    -- create/update SSL certificate file:
    local fname
    fname, msg = updateCRTfile(cfg, cfg.ssl_cert_dir .. "/httpd-default.crt", opts.ssl_sni.cert)
    if fname == nil then
      return false, msg
    end
    fh:write("    SSLCertificateFile ", fname, "\n")

    -- create/update SSL key file:
    fname = cfg.ssl_key_dir .. "/httpd-default.key"
    sfh, msg = io.open(fname, "w")
    if sfh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", fname, "' for writing: ", msg)
      return false, "Can't open '" .. fname .. "' for writing: " .. msg
    end
    sfh:write(opts.ssl_sni.key)
    sfh:close()
    LC.fs.setperm(fname, cfg["ssl_key_mode"], cfg["ssl_key_user"], cfg["ssl_key_group"])
    fh:write("    SSLCertificateKeyFile ", fname, "\n")

    -- update CA chain file
    if opts.ssl_sni.cachain ~= nil then
      if not LC.fs.is_dir(cfg.ssl_chain_dir) then
        LC.fs.mkdir(cfg.ssl_chain_dir)
        LC.fs.setperm(cfg.ssl_chain_dir, 755, "root", "root")
      end
      fname, msg = updateCRTfile(cfg, cfg.ssl_chain_dir .. "/httpd-default-ca.crt", opts.ssl_sni.cachain)
      if fname == false then
        return false, msg
      end
      fh:write("    SSLCertificateChainFile ", fname, "\n")
    end

    fh:write("    <Directory \"", LC.liveconfig.datadir, "/html/\">\n")
    fh:write([[
        <IfModule mod_include.c>
            ErrorDocument 404 /not-available-ssl.shtml
            <IfModule !mod_rewrite.c>
                DirectoryIndex not-available-ssl.shtml
            </IfModule>
        </IfModule>
        <IfModule !mod_include.c>
            ErrorDocument 404 /not-available-ssl.html
            <IfModule !mod_rewrite.c>
                DirectoryIndex not-available-ssl.html
            </IfModule>
        </IfModule>
]])
    fh:write("    </Directory>\n")

    if LC.fs.is_file(vhostpath .. "/default.inc") then
      fh:write("\n    Include ", vhostpath, "/default.inc\n")
    end

    fh:write("</VirtualHost>\n")
    fh:write("</IfModule>\n\n")
  end

  LC.liveconfig.writeFooter(fh)
  fh:close()

  -- rename config file (replacing atomically)
  LC.fs.rename(vhostpath .. "/" .. vhostfile .. ".tmp", vhostpath .. "/" .. vhostfile)

  -- update status file
  LC.liveconfig.writeStatus(statusfile, cfg.type, opts.revision, os.date("%Y-%m-%d %H:%M:%S %Z"))

  -- restart web server (not just reload, so changed interface configuration can be applied)
  LC.timeout.set('apache.restart', RELOAD_MIN, RELOAD_MAX)

  return true
end

-- ---------------------------------------------------------------------------
local function writeVHostIPs(fh, vhost)
  local i, key
  fh:write("<VirtualHost")
  for i, key in ipairs(vhost.ips) do
    if vhost.ssl then
      if string.find(key, ":") then
        -- IPv6 address
        fh:write(" [", key, "]:443")
      else
        -- IPv4 address
        fh:write(" ", key, ":443")
      end
    else
      if string.find(key, ":") then
        -- IPv6 address
        fh:write(" [", key, "]:80")
      else
        -- IPv4 address
        fh:write(" ", key, ":80")
      end
    end
  end
  fh:write(">\n")
end

-- ---------------------------------------------------------------------------
local function writeVHostSSL(fh, cfg, vhost)
  -- SSL configuration:
  fh:write("    SSLEngine On\n")
  local protocols
  if vhost.ssl_strong then
    -- modern/strong cipher suite
    protocols = SSL_PROTOCOLS_STRONG
    fh:write("    SSLCipherSuite ", LC.liveconfig.DEFAULT_SSL_STRONG_CIPHERS, "\n")
  else
    protocols = SSL_PROTOCOLS
    fh:write("    SSLCipherSuite ", LC.liveconfig.DEFAULT_SSL_CIPHERS, "\n")
  end

  if vhost.ssl_notls1 then
    -- disable TLS 1.0 and 1.1
    if not protocols:match('-TLSv1 ') and not protocols:match('-TLSv1$') then
      protocols = protocols .. ' -TLSv1'
    end
    if not protocols:match('-TLSv1%.1 ') and not protocols:match('-TLSv1%.1$') then
      protocols = protocols .. ' -TLSv1.1'
    end
  end
  fh:write("    SSLProtocol ", protocols, "\n")

  -- create/update SSL certificate file:
  local fname, msg
  fname, msg = updateCRTfile(cfg, cfg.ssl_cert_dir .. "/" .. vhost.ssl_filename .. ".crt", vhost.ssl_crt)
  if fname == nil then
    return false, msg
  end
  fh:write("    SSLCertificateFile ", fname, "\n")

  -- create/update SSL key file:
  fname = cfg.ssl_key_dir .. "/" .. vhost.ssl_filename .. ".key"
  sfh, msg = io.open(fname .. ".tmp", "w")
  if sfh == nil then
    LC.log.print(LC.log.ERR, "Can't open '", fname, ".tmp' for writing: ", msg)
    return false, "Can't open '" .. fname .. ".tmp' for writing: " .. msg
  end
  sfh:write(vhost.ssl_key)
  sfh:close()
  LC.fs.setperm(fname .. ".tmp", cfg["ssl_key_mode"], cfg["ssl_key_user"], cfg["ssl_key_group"])
  LC.fs.rename(fname .. ".tmp", fname)
  fh:write("    SSLCertificateKeyFile ", fname, "\n")

  -- create/update SSL CA chain file:
  if vhost.ssl_ca ~= nil then
    if not LC.fs.is_dir(cfg.ssl_chain_dir) then
      LC.fs.mkdir(cfg.ssl_chain_dir)
      LC.fs.setperm(cfg.ssl_chain_dir, 755, "root", "root")
    end
    fname, msg = updateCRTfile(cfg, cfg.ssl_chain_dir .. "/" .. vhost.ssl_filename .. "-ca.crt", vhost.ssl_ca)
    fh:write("    SSLCertificateChainFile ", fname, "\n")
  end

  -- does certificate contain an OCSP URL?
  if vhost.ssl_ocsp then
    -- configure OCSP (with Apache >=2.3.3):
    local v_major, v_minor, v_patch = string.match(cfg.binversion, "^(%d+)%.(%d+)%.(%d+)")
    v_major = tonumber(v_major)
    v_minor = tonumber(v_minor)
    v_patch = tonumber(v_patch)
    if v_major > 2 or (v_major == 2 and (v_minor > 3 or (v_minor == 3 and v_patch >= 3))) then
      fh:write("    SSLUseStapling on\n")
    end
  end

  if cfg.has_http2 and vhost.http2 then
    fh:write("    <IfModule mod_http2.c>\n")
    fh:write("      Protocols h2 http/1.1\n")
    fh:write("    </IfModule>\n")
  end

  fh:write("\n")
end

-- ---------------------------------------------------------------------------
local function writeVHostDocRoot(fh, cfg, opts, vhost, path, php)

  local v_major, v_minor, v_patch = string.match(cfg["binversion"], "^(%d+)%.(%d+)%.(%d+)")
  v_major = tonumber(v_major)
  v_minor = tonumber(v_minor)
  v_patch = tonumber(v_patch)

  -- name of PHP module (mod_php5 or mod_php7) for <IfModule> sections:
  local phpModule = "mod_php5.c"
  if LC.liveconfig.inTable("php7", cfg.modules) then phpModule="mod_php7.c" end

  if php == '' then
    -- get default PHP version code:
    local p
    local phplist = LC.web.getPHP()
    for p in pairs(phplist) do
      if phplist[p].prio == 0 then
        php = phplist[p].code
      end
    end
    -- fallback: "php5" (or "php7")
    if php == '' then php = LC.web.getPHPDefaultCode() end
    -- still not found?
    if php == nil and LC.liveconfig.inTable("php5", cfg.modules) then php="php5" end
    if php == nil and LC.liveconfig.inTable("php7", cfg.modules) then php="php7" end
  end

  if opts.path then
    -- we have a DocumentRoot
    fh:write("    DocumentRoot \"", escape(path), "\"\n")
    fh:write("    <Directory \"", escape(path), "\">\n")
    fh:write("        Options MultiViews")
    if FOLLOWSYMLINKS then
      fh:write(" FollowSymLinks")
    else
      fh:write(" SymLinksIfOwnerMatch")
    end
    if opts.hasCGI and opts.hasSSI then
      fh:write(" Includes")
    elseif opts.hasSSI then
      fh:write(" IncludesNOEXEC")
    end
    fh:write("\n")
    fh:write("        AllowOverride AuthConfig Indexes Limit")
    if #RESTRICT_FILEINFO == 0 or v_major < 2 or (v_major == 2 and v_minor < 3) then
      -- AllowOverride FileInfo is only allowed with Apache <2.3
      fh:write(" FileInfo")
    end
    fh:write(" Options=Indexes,MultiViews")
    if FOLLOWSYMLINKS then
      fh:write(",FollowSymLinks")
    else
      fh:write(",SymLinksIfOwnerMatch")
    end
    if opts.hasCGI then
      fh:write(",ExecCGI")
      if opts.hasSSI then
        fh:write(",Includes")
      end
    elseif opts.hasSSI then
      fh:write(",IncludesNOEXEC")
    end
    fh:write("\n")
    if (v_major > 2 or (v_major == 2 and v_minor >= 3)) and #RESTRICT_FILEINFO > 0 then
      -- restricted commands in .htaccess
      fh:write("        AllowOverrideList")
      local idx, value
      for idx, value in pairs(RESTRICT_FILEINFO) do
        fh:write(" ", value)
      end
      fh:write("\n")
    end
    if v_major < 2 or (v_major == 2 and v_minor < 3) then
      fh:write("        Order allow,deny\n")
      fh:write("        allow from all\n")
    else
      fh:write("        <IfModule mod_access_compat.c>\n")
      fh:write("            Order allow,deny\n")
      fh:write("            Allow from all\n")
      fh:write("        </IfModule>\n")
      fh:write("        <IfModule mod_authz_core.c>\n")
      fh:write("            Require all granted\n")
      fh:write("        </IfModule>\n")
    end
    fh:write("    </Directory>\n\n")
  end

  -- error.log enabled?
  if opts.errorlog then
    fh:write("    ErrorLog \"", opts.path, "/logs/error.log\"\n\n")
    if not LC.fs.is_file(opts.path .. "/logs/error.log") then
      -- touch file
      local fh, msg = io.open(opts.path .. "/logs/error.log", "w")
      fh:close()
    end
    -- make error.log readable. the 'logs' directory itself is protected against unauthorized access
    LC.fs.setperm(opts.path .. "/logs/error.log", 644, "root", "root")
  end

  -- CGI allowed?
  if opts.hasCGI then
    -- check if cgi-bin exists, eventually create it:
    if not LC.fs.is_dir(opts.path .. "/" .. LC.web.HTDOCS_PATH .. "/cgi-bin") then
      LC.fs.mkdir(opts.path .. "/" .. LC.web.HTDOCS_PATH .. "/cgi-bin")
      LC.fs.setperm(opts.path .. "/" .. LC.web.HTDOCS_PATH .. "/cgi-bin", 755, opts.user, opts.group)
      if LC.distribution.family == "RedHat" then
        LC.exec("chcon --type=user_home_t \"" .. opts.path .. "/cgi-bin\"")
      end
    end
    fh:write("    ScriptAlias /cgi-bin/ \"", opts.path, "/" .. LC.web.HTDOCS_PATH .. "/cgi-bin/\"\n")
    fh:write("    <Directory \"", opts.path, "/" .. LC.web.HTDOCS_PATH .. "/cgi-bin/\">\n")
    fh:write("        Options -MultiViews\n")
    fh:write("    </Directory>\n\n")

    -- configure suExec User/Group
    if opts.user ~= nil and opts.group ~= nil then
      fh:write("    <IfModule mod_suexec.c>\n")
      fh:write("        SuexecUserGroup ", opts.user, " ", opts.group, "\n")
      fh:write("    </IfModule>\n\n")
    end
  else
    -- forbid access to /cgi-bin/ directory!
    fh:write("    <Directory \"", opts.path, "/" .. LC.web.HTDOCS_PATH .. "/cgi-bin\">\n")
    if v_major < 2 or (v_major == 2 and v_minor < 3) then
      fh:write("        Order allow,deny\n")
      fh:write("        deny from all\n")
    else
      fh:write("        <IfModule mod_access_compat.c>\n")
      fh:write("            Order allow,deny\n")
      fh:write("            deny from all\n")
      fh:write("        </IfModule>\n")
      fh:write("        <IfModule mod_authz_core.c>\n")
      fh:write("            Require all denied\n")
      fh:write("        </IfModule>\n")
    end
    fh:write("    </Directory>\n\n")
  end

  -- configure for mpm_itk?
  if LC.liveconfig.inTable("mpm_itk", cfg.modules) then
    fh:write("    <IfModule mpm_itk_module>\n")
    fh:write("        AssignUserID ", opts.user, " ", opts.group, "\n")
    fh:write("    </IfModule>\n\n")
  end

  -- configure tmp directory
  if opts.hasCGI or opts.hasPHP or opts.hasSSI then
    fh:write("    <IfModule mod_env.c>\n")
    fh:write("        SetEnv TMP \"", opts.path, "/tmp\"\n")
    fh:write("        SetEnv TMPDIR \"", opts.path, "/tmp\"\n")
    fh:write("    </IfModule>\n\n")
  end

  -- PHP settings
  if opts.hasPHP and opts.hasPHP > 0 then
    -- PHP enabled:
    local phpmode = { "suPHP", "FastCGI", "mod_php", "FPM" }
    fh:write("    # PHP configuration for this subscription: ")
    if opts.hasPHP == 1 then
      local handler = "application/x-httpd-suphp"
      if php ~= LC.web.getPHPDefaultCode() then
        handler = "application/x-httpd-" .. php
      elseif LC.distribution.family == "RedHat" then
        handler = "x-httpd-php"
      elseif LC.distribution.family == "SUSE" then
        handler = "application/x-httpd-php"
      end
      fh:write("suPHP\n")
      -- suPHP configuration
      fh:write("    <IfModule mod_suphp.c>\n")
      fh:write("        <IfModule ", phpModule, ">\n")
      fh:write("            php_admin_flag engine off\n")
      fh:write("            <FilesMatch \"\\.ph(p3?|tml)$\">\n")
      fh:write("                SetHandler ", handler, "\n")
      fh:write("            </FilesMatch>\n")
        -- FilesMatch for Debian 7:
        fh:write("            <FilesMatch \".+\\.ph(p[3457]?|t|tml|ps)$\">\n")
        fh:write("                SetHandler ", handler, "\n")
        fh:write("            </FilesMatch>\n")
      if LC.distribution.family == "SUSE"  then
        fh:write("            <FilesMatch \"\\.ph(p[345]?|tml)$\">\n")
        fh:write("                SetHandler ", handler, "\n")
        fh:write("            </FilesMatch>\n")
      end
      fh:write("        </IfModule>\n")
      fh:write("        suPHP_Engine on\n")
      if LC.distribution.family == "RedHat"  then
        fh:write("        suPHP_AddHandler x-httpd-php\n")
      end
      -- add suPHP_User and suPHP_Group for RHEL/CentOS:
      if LC.distribution.family == "RedHat" or LC.distribution.family == "SUSE" then
        if opts.user ~= nil and opts.group ~= nil then
          fh:write("        suPHP_UserGroup ", opts.user, " ", opts.group, "\n")
        end
      end
      -- use another PHP version?
      if php ~= LC.web.getPHPDefaultCode() then
        fh:write("        suPHP_AddHandler application/x-httpd-" .. php .. "\n")
        fh:write("        AddType application/x-httpd-" .. php .. " .php .php3 .php4 .php5 .phtml\n")
      end
      if LC.fs.is_file(opts.path .. "/.php5/php.ini") then
        -- use custom php.ini file
        fh:write("        suPHP_ConfigPath ", opts.path, "/.php5/\n")
      elseif LC.fs.is_file(opts.path .. "/conf/" .. php .. "/php.ini") then
        fh:write("        suPHP_ConfigPath ", opts.path, "/conf/" , php , "/\n")
      end
      fh:write("    </IfModule>\n\n")

    elseif opts.hasPHP == 2 then
      fh:write("FastCGI\n")
      -- FastCGI configuration
      fh:write("    <IfModule mod_fcgid.c>\n")
      fh:write("        <IfModule ", phpModule, ">\n")
      fh:write("            php_admin_flag engine off\n")
      fh:write("            <FilesMatch \"\\.ph(p3?|tml)$\">\n")
      fh:write("                SetHandler None\n")
      fh:write("            </FilesMatch>\n")
        -- FilesMatch for Debian 7:
        fh:write("            <FilesMatch \".+\\.ph(p[3457]?|t|tml|ps)$\">\n")
        fh:write("                SetHandler None\n")
        fh:write("            </FilesMatch>\n")
      fh:write("        </IfModule>\n")
      if LC.distribution.family == "SUSE"  then
        fh:write("            <FilesMatch \"\\.ph(p[345]?|tml)$\">\n")
        fh:write("                SetHandler None\n")
        fh:write("            </FilesMatch>\n")
      end
      if not opts.hasCGI then
        -- if not already defined, set suexecUserGroup here:
        fh:write("        <IfModule mod_suexec.c>\n")
        fh:write("            SuexecUserGroup ", opts.user, " ", opts.group, "\n")
        fh:write("        </IfModule>\n\n")
      end
      fh:write("        <FilesMatch \"\\.php5?$\">\n")
      fh:write("            Options +ExecCGI\n")
      fh:write("            SetHandler fcgid-script\n")
      fh:write("        </FilesMatch>\n")
      fh:write("        FcgidWrapper ", opts.path, "/conf/", php, "/php-fcgi-starter .php\n")
      fh:write("        FcgidWrapper ", opts.path, "/conf/", php, "/php-fcgi-starter .php5\n")
      if opts.phpini and opts.phpini['post_max_size'] then
        local val = LC.web.php2int(opts.phpini['post_max_size'].value)
        fh:write("        FcgidMaxRequestLen ", val, "\n")
      end
      if opts.phpini and (opts.phpini['max_execution_time'] or opts.phpini['max_input_time']) then
        local timeout = 40
        if opts.phpini['max_execution_time'] then
          local val = LC.web.php2int(opts.phpini['max_execution_time'].value)
          if timeout < val then timeout = val end
        end
        if opts.phpini['max_input_time'] then
          local val = LC.web.php2int(opts.phpini['max_input_time'].value)
          if timeout < val then timeout = val end
        end
        fh:write("        FcgidBusyTimeout ", timeout, "\n")
        fh:write("        FcgidIOTimeout ", timeout, "\n")
      end

      fh:write("        FcgidMaxRequestsPerProcess 5000\n")
      fh:write("    </IfModule>\n")

    elseif opts.hasPHP == 3 then
      fh:write("mod_php\n")
      -- mod_php configuration
      fh:write("    <IfModule ", phpModule, ">\n")
      if LC.distribution.family == "Debian" then
        -- re-add handler for application/x-httpd (eg. if suPHP is also available)
        fh:write("        <FilesMatch \".+\\.ph(p[3457]?|t|tml|ps)$\">\n")
        fh:write("            SetHandler application/x-httpd-php\n")
        fh:write("        </FilesMatch>\n")
      elseif LC.distribution.family == "Redhat" then
        fh:write("        AddHandler php5-script .php\n")
      end
      local pv = LCS.web.phpVersionInt    -- PHP version (as integer, eg. "50411")
      local i, value, v_min, v_max
      for i in pairs(opts.phpini) do
        value = opts.phpini[i].value
        value = string.gsub(value, "%%HOME%%", opts.path)
        value = string.gsub(value, "%%PHP%%", php)
        if opts.phpini[i].t == 'e' then
          -- value is an expression -> this must be evaluated first!
          -- cache evaluated statements in local table "phpEval"
          -- "Sane" expressions (coming from LiveConfig) may only contain upper-case characters (no function calls etc.).
          -- Nevertheless, we're "paranoid" here to minimize impact when someone had access to the LiveConfig database.
          if phpEval[value] == nil and not string.match(value, "[a-z'`;]") then
            local ph = io.popen("/usr/bin/php -d disabled_functions='exec,passthru,system,shell_exec,popen,proc_open,pcntl_exec,eval,ssert,preg_replace,create_function,include,include_once,require,require_once,phpinfo,putenv,ini_set,pfsockopen,fsockopen,open' -d open_basedir=/tmp -r 'echo " .. value .. ";'")
            if ph ~= nil then
              phpEval[value] = ph:read()
              ph:close()
            end
          end
          if phpEval[value] == nil then
            -- comment-out this value!
            fh:write("#")
          else
            value = phpEval[value]
          end
        end
        v_min = opts.phpini[i].min  -- minimum supported PHP version (as integer)
        v_max = opts.phpini[i].max  -- supported up to this PHP version (not included)
        if pv == nil or ((v_min == nil or pv >= v_min) and (v_max == nil or pv < v_max)) then
          fh:write("        php_")
          if opts.phpini[i].o == nil or opts.phpini[i].o == false then
            fh:write("admin_")
          end
          if opts.phpini[i].t and opts.phpini[i].t == 'b' then
            fh:write("flag")
          else
            fh:write("value")
          end
          fh:write(" ", i, " ", value, "\n")
        end
      end
      fh:write("    </IfModule>\n\n")
    elseif opts.hasPHP == 4 then
      fh:write("FPM\n")
      -- FPM configuration
      fh:write("    <IfModule ", phpModule, ">\n")
      fh:write("        php_admin_flag engine off\n")
      fh:write("    </IfModule>\n")
      fh:write("    <IfModule mod_proxy_fcgi.c>\n")
      fh:write("        <FilesMatch \".+\\.ph(p[3457]?|t|tml|ps)$\">\n")
      local p, found
      for p in pairs(LCS.web.phpVersions) do
        if LCS.web.phpVersions[p].code == php then
          local phpconfig = LCS.web.phpVersions[p]
          if phpconfig.fpm == nil then break end
          found = true
          -- opts.name:lower()! the proxy destination URL is converted to lower-case by Apache!!
          -- so the .sock location must be completely lowercase on the server!
          fh:write("            SetHandler \"proxy:unix:", phpconfig.fpm.sockets, "/", opts.name:lower(), ".sock|fcgi://localhost\"\n")
          fh:write("            SetEnvIfNoCase ^Authorization$ \"(.+)\" HTTP_AUTHORIZATION=$1\n")
          break
        end
      end
      if not found then
        LC.log.print(LC.log.ERR, "Apache: PHP version ", pv, " configured as FPM for subscription ", opts.name, " but no PHP found for this version!")
        -- to prevent raw .php files being served, return error 502 (Bad Gateway):
        fh:write("            Redirect 502 /\n")
      end
      fh:write("        </FilesMatch>\n")
      if not found then
        -- ErrorDocument must be outside of FilesMatch section
        fh:write("        ErrorDocument 502 \"Configuration error: FPM configured for this domain, but no matching PHP-FPM found. Please contact your administrator.\"\n")
      end
      fh:write("    </IfModule>\n\n")
    else
      fh:write("-unknown- (", opts.hasPHP, ")\n")
    end

  else
    -- PHP disabled:

    -- suPHP configuration
    fh:write("    <IfModule mod_suphp.c>\n")
    fh:write("        <Directory \"", opts.path, "\">\n")
    fh:write("            suPHP_Engine Off\n")
    fh:write("        </Directory>\n")
    fh:write("    </IfModule>\n\n")

    -- mod_php configuration
    fh:write("    <IfModule ", phpModule, ">\n")
    fh:write("        php_admin_flag engine Off\n")
    fh:write("    </IfModule>\n\n")
  end

  if opts.pwprotect ~= nil then
    local pwpath
    fh:write("    # password-protected directories:\n")
    for pwpath in pairs(opts.pwprotect) do
      fh:write("    <Directory \"", opts.path, escape(pwpath), "\">\n")
      fh:write("        AuthType Basic\n")
      fh:write("        AuthUserFile ", opts.path, "/conf/.htpasswd\n")
      if v_major > 2 or (v_major == 2 and v_minor >= 1) then
        fh:write("        <IfModule mod_authz_groupfile.c>\n")
        fh:write("            AuthGroupFile /dev/null\n")
        fh:write("        </IfModule>\n")
      end
      fh:write("        AuthName \"", escape(opts.pwprotect[pwpath].title), "\"\n")
      if opts.pwprotect[pwpath].users[1] == "*" then
        fh:write("        Require valid-user\n")
      else
        fh:write("        Require user")
        local i
        for i in pairs(opts.pwprotect[pwpath].users) do
          fh:write(" \"", escape(opts.pwprotect[pwpath].users[i]), "\"")
        end
        fh:write("\n")
      end
      fh:write("    </Directory>\n")
    end
    fh:write("\n")
  end
end

-- ---------------------------------------------------------------------------
local function writeVHostStats(fh, cfg, opts)
  local v_major, v_minor, v_patch = string.match(cfg["binversion"], "^(%d+)%.(%d+)%.(%d+)")
  v_major = tonumber(v_major)
  v_minor = tonumber(v_minor)
  v_patch = tonumber(v_patch)

  fh:write("        # Web statistics:\n")
  if opts.webstats.software == 1 then
    -- shared resources for Webalizer
    fh:write("        RewriteCond %{HTTP_HOST} ^", string.gsub(opts.webstats.domain, "%.", "\\."), "\\.?$ [NC]\n")
    fh:write("        RewriteRule ^/", string.gsub(opts.webstats.path, "%.", "\\."), "/images/(.*) /usr/share/liveconfig/html/webalizer/$1 [L]\n")
  elseif opts.webstats.software == 2 then
    -- shared resources for AWStats
    local icondir = "/usr/share/awstats/icon"
    if LC.distribution.family == "RedHat" then
      icondir = "/var/www/awstats/icon"
    elseif LC.distribution.family == "SUSE" then
      icondir = "/usr/share/awstats/icon"
    elseif LC.distribution.family == "Gentoo" then
      icondir = "/usr/share/awstats/wwwroot/icon"
    end
    fh:write("        RewriteCond %{HTTP_HOST} ^", string.gsub(opts.webstats.domain, "%.", "\\."), "\\.?$ [NC]\n")
    fh:write("        RewriteRule ^/", string.gsub(opts.webstats.path, "%.", "\\."), "/icon/(.*) ", icondir, "/$1 [L]\n")
    fh:write("        RewriteCond %{HTTP_HOST} ^", string.gsub(opts.webstats.domain, "%.", "\\."), "\\.?$ [NC]\n")
    fh:write("        RewriteRule ^/", string.gsub(opts.webstats.path, "%.", "\\."), "/awstats/(.*) /usr/share/liveconfig/html/awstats/$1 [L]\n")
  end
  fh:write("        RewriteCond %{HTTP_HOST} ^", string.gsub(opts.webstats.domain, "%.", "\\."), "\\.?$ [NC]\n")
  fh:write("        RewriteRule ^/", string.gsub(opts.webstats.path, "%.", "\\."), "/(.*) ", opts.path, "/stats/$1 [L]\n")
  fh:write("        <Directory \"" .. opts.path .. "/stats/\">\n")
  fh:write("            AuthType Basic\n")
  fh:write("            AuthName \"Web Statistics\"\n")
  fh:write("            AuthUserFile \"" .. opts.path .. "/stats/.htpasswd\"\n")
  if v_major > 2 or (v_major == 2 and v_minor >= 1) then
    fh:write("            <IfModule mod_authz_groupfile.c>\n")
    fh:write("                AuthGroupFile /dev/null\n")
    fh:write("            </IfModule>\n")
  end
  if v_major > 2 or (v_major == 2 and v_minor >= 3) then
    fh:write("            <IfModule mod_access_compat.c>\n")
  end
  fh:write("            Order allow,deny\n")
  fh:write("            allow from all\n")
  if v_major > 2 or (v_major == 2 and v_minor >= 3) then
    fh:write("            </IfModule>\n")
  end
  fh:write("            Require valid-user\n")
  fh:write("        </Directory>\n\n")

end

-- ---------------------------------------------------------------------------
local function writeVHostRedirects(fh, r301, r302)
  -- create RewriteRules for all domains with 301-redirects
  for key in pairs(r301) do
    -- fh:write("... ", key, "...\n")
    if key ~= "" then
      local d
      for i, d in ipairs(r301[key]) do
        d = string.gsub(d, "%.", "\\.")
        d = string.gsub(d, "%*\\%.", "(.+\\.)*")
        fh:write("        RewriteCond %{HTTP_HOST} ^", d, "\\.?$ [NC")
        if i < #r301[key] then fh:write(",OR") end
        fh:write("]\n")
      end
      if string.sub(key, 1, 9) == 'https://*' then
        key = 'https://%{HTTP_HOST}/*'
      else
        key = string.gsub(key, "\\", "\\\\")
        key = string.gsub(key, " ", "\\ ")
        key = string.gsub(key, "%$", "\\$")
        key = string.gsub(key, "(%%{[^}]*})", "\\%1")
      end
      if string.sub(key, -2) == '/*' then
        fh:write("        RewriteRule ^(.*)$      ", string.sub(key, 1, string.len(key)-2), "$1 [R=301,NE,L]\n\n")
      else
        fh:write("        RewriteRule ^.*         ", key, " [R=301,NE,L]\n\n")
      end
    end
  end

  -- create RewriteRules for all domains with 302-redirects
  for key in pairs(r302) do
    -- fh:write("... ", key, "...\n")
    if key ~= "" then
      local d
      for i, d in ipairs(r302[key]) do
        d = string.gsub(d, "%.", "\\.")
        d = string.gsub(d, "%*\\%.", "(.+\\.)*")
        fh:write("        RewriteCond %{HTTP_HOST} ^", d, "\\.?$ [NC")
        if i < #r302[key] then fh:write(",OR") end
        fh:write("]\n")
      end
      if string.sub(key, 1, 9) == 'https://*' then
        key = 'https://%{HTTP_HOST}/*'
      else
        key = string.gsub(key, "\\", "\\\\")
        key = string.gsub(key, " ", "\\ ")
        key = string.gsub(key, "%$", "\\$")
        key = string.gsub(key, "(%%{[^}]*})", "\\%1")
      end
      if string.sub(key, -2) == '/*' then
        fh:write("        RewriteRule ^(.*)$      ", string.sub(key, 1, string.len(key)-2), "$1 [R=302,NE,L]\n\n")
      else
        fh:write("        RewriteRule ^.*         ", key, " [R=302,NE,L]\n\n")
      end
    end
  end

end

-- ---------------------------------------------------------------------------
local function writeACME(fh, cfg, opts, vhost)
  -- write ACME challenges
  -- only for backward-compatibility with "old" Let's Encrypt / ACME interface!

  local key
  for key in pairs(vhost.domains) do
    if vhost.domains[key].acme ~= nil then
      local acme = vhost.domains[key].acme
      local idx
      for idx = 1, #acme do
        LC.web.addValidation(key, "/acme-challenge/", acme[idx].token, acme[idx].response)
      end
    end
  end
end

-- ---------------------------------------------------------------------------
local function writeInclude(fh, path, domainInclude)
  -- include custom configuration options
  if LC.fs.is_file(path .. "/conf/httpd.conf") then
    fh:write("    # Include customer-specific configuration options:\n")
    fh:write("    Include ", path, "/conf/httpd.conf\n\n")
  elseif LC.fs.is_file(path .. "/.httpd.conf") then
    -- old ("legacy") location:
    fh:write("    # Include customer-specific configuration options:\n")
    fh:write("    Include ", path, "/.httpd.conf\n\n")
  end

  if domainInclude ~= nil and LC.fs.is_file(domainInclude) then
    fh:write("    # Include customer-specific per-domain configuration options:\n")
    fh:write("    Include ", domainInclude, "\n\n")
  end
end

-- ---------------------------------------------------------------------------
-- Build <VirtualHost> sections (called by configureVHost())
-- ---------------------------------------------------------------------------
local function buildConfig(fh, cfg, opts, vhost, domains)
  local cnames = { }
  local v_major, v_minor, v_patch = string.match(cfg.binversion, "^(%d+)%.(%d+)%.(%d+)")
  v_major = tonumber(v_major)
  v_minor = tonumber(v_minor)
  v_patch = tonumber(v_patch)
  -- copy all domain names in a table, and then sort them:
  local sd   = { }  -- list with all domains/subdomains and their destination directory, grouped by PHP version!
  local sd_count = 0
  local r301 = { }  -- list with all domains/subdomains with 301-redirect
  local r301_count = 0
  local r302 = { }  -- list with all domains/subdomains with 302-redirect
  local r302_count = 0
  local prx  = { }  -- list with all domains/subdomains with proxy destination
  local prx_count = 0   -- Lua doesn't support counting table entries with non-integer keys :(
  local app  = { }  -- list with all domains/subdomains with web application, grouped by PHP version!
  local app_count = 0
  local bldr = { }  -- list with all domains/subdomains with website builder destination
  local bldr_count = 0
  local idx, domain, i, key, dst
  for idx, domain in pairs(domains) do

    dst = vhost.domains[domain].dest
    if dst == nil then dst = '' end

    -- insert pseudo path "/#/<domain>" into docroot for per-domain httpd.conf includes
    -- (except if domain is redirected)
    if vhost.domains[domain].type ~= 2 and vhost.domains[domain].type ~= 3 and LC.fs.is_file(opts.path .. "/conf/" .. domain .. ".httpd.conf") then
      dst = dst .. "/#/" .. domain
    end

    local php = vhost.domains[domain].php
    if php == nil then php = '' end

    if vhost.domains[domain].type == 1 then
      -- normal (local) (sub)domain; add to subdomain list, grouped by PHP version code
      if sd[php] == nil then sd[php] = { } end
      if sd[php][dst] == nil then sd[php][dst] = { } end
      sd[php][dst][#sd[php][dst] + 1] = domain
      sd_count = sd_count + 1
    end

    if vhost.domains[domain].type == 2 then
      -- (sub)domain with 301-redirect
      if r301[dst] == nil then r301[dst] = { } end
      r301[dst][#r301[dst] + 1] = domain
      r301_count = r301_count + 1
    end

    if vhost.domains[domain].type == 3 then
      -- (sub)domain with 302-redirect
      if r302[dst] == nil then r302[dst] = { } end
      r302[dst][#r302[dst] + 1] = domain
      r302_count = r302_count + 1
    end

    if vhost.domains[domain].type == 4 then
      -- (sub)domain with proxy destination
      if prx[dst] == nil then prx[dst] = { } end
      prx[dst][#prx[dst] + 1] = domain
      prx_count = prx_count + 1
    end

    if vhost.domains[domain].type == 5 then
      -- (sub)domain with web application, grouped by PHP version code!
      if app[php] == nil then app[php] = { } end
      if app[php][dst] == nil then app[php][dst] = { } end
      app[php][dst][#app[php][dst] + 1] = domain
      app_count = app_count + 1
    end

    if vhost.domains[domain].type == 7 then
      -- Website builder
      if bldr[dst] == nil then bldr[dst] = { } end
      bldr[dst][#bldr[dst] + 1] = domain
      bldr_count = bldr_count + 1
    end

  end

  if string.match(domains[1], "^%*%.") then
    domains[#domains + 1] = domains[1]
    domains[1] = "_catchall_." .. string.sub(domains[1], 3)
  end

  if vhost.ssl then
    fh:write("<IfModule mod_ssl.c>\n")
  end

  if opts.suspended then
    -- website has been suspended
    writeVHostIPs(fh, vhost)

    for idx, domain in ipairs(domains) do
      if idx == 1 then
        fh:write("    ServerName ", domain, "\n")
        cnames[#cnames+1] = domain
      else
        fh:write("    ServerAlias ", domain, "\n")
      end
    end
    fh:write("\n")

    if vhost.ssl then
      writeVHostSSL(fh, cfg, vhost)
    end

    fh:write("    <IfModule mod_rewrite.c>\n")
    fh:write("        RewriteEngine On\n")
    fh:write("        RewriteRule ^/\\.errorFiles/.* - [L]\n")
    fh:write("        RewriteRule .* /.errorFiles/suspended.html [R=503,L]\n")
    fh:write("    </IfModule>\n\n")
    fh:write("    ErrorDocument 503 /.errorFiles/suspended.html\n\n")
    fh:write("</VirtualHost>\n")

  elseif CONFIGMODE == 'rewrite' then
    -- Configure all domains/subdomains of a subscription within one single <VirtualHost>
    -- statement. This provides best performance when managing a very large number of
    -- subscriptions and domains on a single machine.
    writeVHostIPs(fh, vhost)

    local php
    for idx, domain in ipairs(domains) do
      if idx == 1 then
        fh:write("    ServerName ", domain, "\n")
        cnames[#cnames+1] = domain
        php = vhost.domains[domain].php
      else
        fh:write("    ServerAlias ", domain, "\n")
      end
    end
    if php == nil then php = '' end
    fh:write("\n")

    if vhost.ssl then
      writeVHostSSL(fh, cfg, vhost)
    end

    if opts.path then
      writeVHostDocRoot(fh, cfg, opts, vhost, opts.path .. "/" .. LC.web.HTDOCS_PATH, php)
    end

    -- write ACME challenges
    writeACME(fh, cfg, opts, vhost)

    -- enable RewriteEngine
    fh:write("    <IfModule mod_rewrite.c>\n")
    fh:write("        RewriteEngine On\n\n")

    -- /.well-known/ hook for domain validation (SSL certificates):
    fh:write("        RewriteMap  \"lc\" \"int:tolower\"\n")
    fh:write("        RewriteCond %{REQUEST_URI} ^/\\.well-known/(.*/)?([^/]*)$\n")
    fh:write("        RewriteCond ", opts.path .. "/" .. LC.web.HTDOCS_PATH, "/.well-known/%1%2 !-f\n")
    fh:write("        RewriteCond ", LC.web.getWebRoot(), "/.well-known/htdocs/%1${lc:%{HTTP_HOST}}.%2 -f\n")
    fh:write("        RewriteRule / ", LC.web.getWebRoot(), "/.well-known/htdocs/%1${lc:%{HTTP_HOST}}.%2 [T=text/plain,L]\n")

    if opts.hasCGI then
      -- don't rewrite CGI URLs (if user has CGI permission)
      if LC.distribution.family == "SUSE" then
        -- SUSE has its own /error/ alias - keep it untouched:
        fh:write("        # Don't rewrite /cgi-bin/, /error/, /icons/ and /.errorFiles/ URLs:\n")
        fh:write("        RewriteRule ^/(cgi-bin|error|icons|\\.errorFiles)/.* - [L]\n\n")
      else
        fh:write("        # Don't rewrite /cgi-bin/, /icons/ and /.errorFiles/ URLs:\n")
        fh:write("        RewriteRule ^/(cgi-bin|icons|\\.errorFiles)/.* - [L]\n\n")
      end
    else
      if LC.distribution.family == "SUSE" then
        -- SUSE has its own /error/ alias - keep it untouched:
        fh:write("        # Don't rewrite /error/, /icons/ and /.errorFiles/ URLs:\n")
        fh:write("        RewriteRule ^/(error|icons|\\.errorFiles)/.* - [L]\n\n")
      else
        fh:write("        # Don't rewrite /icons/ and /.errorFiles/ URLs:\n")
        fh:write("        RewriteRule ^/(icons|\\.errorFiles)/.* - [L]\n\n")
      end
    end

    -- Web statistics:
    if opts.webstats and opts.webstats.domain and opts.webstats.path and vhost.domains[opts.webstats.domain] ~= nil then
      writeVHostStats(fh, cfg, opts)
    end

    -- create RewriteRules for all domains pointing to a path within DocumentRoot
    for php in pairs(sd) do
      for key in pairs(sd[php]) do
        -- fh:write("... ", key, "...\n")
        if key ~= "" then
          local d
          for i, d in ipairs(sd[php][key]) do
            d = string.gsub(d, "%.", "\\.")
            d = string.gsub(d, "%*\\%.", "(.+\\.)*")
            fh:write("        RewriteCond %{HTTP_HOST} ^", d, "\\.?$ [NC")
            if i < #sd[php][key] then fh:write(",OR") end
            fh:write("]\n")
          end
          fh:write("        RewriteRule ^/(.*)       ", string.gsub(opts.path .. "/" .. LC.web.HTDOCS_PATH .. "/" .. key, " ", "\\ "), "/$1 [L]\n\n")
        end
      end
    end

    writeVHostRedirects(fh, r301, r302)

    -- create ProxyPass commands for all domains with proxy destinations
    if prx_count > 0 then
      local ssl_dest = false
      fh:write("      <IfModule mod_proxy.c>\n\n")
      for key in pairs(prx) do
        -- fh:write("... ", key, "...\n")
        if key ~= "" then
          local d
          for i, d in ipairs(prx[key]) do
            d = string.gsub(d, "%.", "\\.")
            d = string.gsub(d, "%*\\%.", "(.+\\.)*")
            fh:write("        RewriteCond %{HTTP_HOST} ^", d, "\\.?$ [NC")
            if i < #prx[key] then fh:write(",OR") end
            fh:write("]\n")
          end
          d = key
          if string.sub(d, -2) == '/*' then d = string.sub(d, 1, string.len(d)-2) end
          fh:write("        RewriteRule ^(.*)        ", string.gsub(d, " ", "\\ "), "$1 [P,L]\n\n")
          if string.match(key, "^https://") then ssl_dest = true end
        end
      end
      if ssl_dest then
        fh:write("        <IfModule mod_ssl.c>\n")
        fh:write("          SSLProxyEngine On\n")
        fh:write("        </IfModule>\n")
      end
      fh:write("      </IfModule>\n\n")
    end

    -- create RewriteRules for all domains pointing to a web application:
    if app_count > 0 then
      -- allow access to apps directories:
      fh:write("        <Directory \"", opts.path, "/apps\">\n")
      fh:write("            Options MultiViews")
      if FOLLOWSYMLINKS then
        fh:write(" FollowSymLinks")
      else
        fh:write(" SymLinksIfOwnerMatch")
      end
      if opts.hasCGI and opts.hasSSI then
        fh:write(" Includes")
      elseif opts.hasSSI then
        fh:write(" IncludesNOEXEC")
      end
      fh:write("\n")
      fh:write("            AllowOverride AuthConfig Indexes Limit")
      if #RESTRICT_FILEINFO == 0 or v_major < 2 or (v_major == 2 and v_minor < 3) then
        -- AllowOverride FileInfo is only allowed with Apache <2.3
        fh:write(" FileInfo")
      end
      fh:write(" Options=Indexes,MultiViews")
      if FOLLOWSYMLINKS then
        fh:write(",FollowSymLinks")
      else
        fh:write(",SymLinksIfOwnerMatch")
      end
      if opts.hasCGI then
        fh:write(",ExecCGI")
        if opts.hasSSI then
          fh:write(",Includes")
        end
      elseif opts.hasSSI then
        fh:write(",IncludesNOEXEC")
      end
      fh:write("\n")
      if (v_major > 2 or (v_major == 2 and v_minor >= 3)) and #RESTRICT_FILEINFO > 0 then
        -- restricted commands in .htaccess
        fh:write("        AllowOverrideList")
        local idx, value
        for idx, value in pairs(RESTRICT_FILEINFO) do
          fh:write(" ", value)
        end
        fh:write("\n")
      end
      if v_major < 2 or (v_major == 2 and v_minor < 3) then
        fh:write("            Order allow,deny\n")
        fh:write("            allow from all\n")
      else
        fh:write("            <IfModule mod_access_compat.c>\n")
        fh:write("                Order allow,deny\n")
        fh:write("                Allow from all\n")
        fh:write("            </IfModule>\n")
        fh:write("            <IfModule mod_authz_core.c>\n")
        fh:write("                Require all granted\n")
        fh:write("            </IfModule>\n")
      end

      fh:write("        </Directory>\n\n")

      -- add RewriteRules for apps:
      for php in pairs(app) do
        for key in pairs(app[php]) do
          -- fh:write("... ", key, "...\n")
          if key ~= "" then
            local d
            for i, d in ipairs(app[php][key]) do
              d = string.gsub(d, "%.", "\\.")
              d = string.gsub(d, "%*\\%.", "(.+\\.)*")
              fh:write("        RewriteCond %{HTTP_HOST} ^", d, "\\.?$ [NC")
              if i < #app[php][key] then fh:write(",OR") end
              fh:write("]\n")
            end
            fh:write("        RewriteRule ^/(.*)       ", string.gsub(opts.path .. "/apps/" .. key, " ", "\\ "), "/$1 [L]\n\n")
          end
        end
      end
    end

    fh:write("    </IfModule>\n\n")

    -- add support for nice default error pages
    fh:write([[
    <LocationMatch "^/$">
        ErrorDocument 403 /.errorFiles/coming-soon.html
        ErrorDocument 404 /.errorFiles/coming-soon.html
    </LocationMatch>

]])

    -- include custom configuration options
    if LC.fs.is_file(opts.path .. "/.httpd.conf") then
      fh:write("    # Include customer-specific configuration options:\n")
      fh:write("    Include ", opts.path, "/.httpd.conf\n\n")
    end

    fh:write("</VirtualHost>\n")

  elseif CONFIGMODE == 'virtualhost' then
    -- Configure each different domain/subdomain with own <VirtualHost> statement
    -- This provides best compatibility, especially regarding usage of the
    -- DOCUMENT_ROOT environment variable in PHP/CGI scripts

    -- Step 1: configure all local webspace domains:
    if sd_count > 0 then
      for php in pairs(sd) do
        for i in pairs(sd[php]) do
          writeVHostIPs(fh, vhost)

          table.sort(sd[php][i])
          if string.match(sd[php][i][1], "^%*%.") then
            sd[php][i][#sd[php][i] + 1] = sd[php][i][1]
            sd[php][i][1] = "_catchall_." .. string.sub(sd[php][i][1], 3)
          end

          local d
          for d in pairs(sd[php][i]) do
            if d == 1 then
              fh:write("    ServerName ", sd[php][i][d], "\n")
              cnames[#cnames+1] = sd[php][i][d]
            else
              fh:write("    ServerAlias ", sd[php][i][d], "\n")
            end
          end
          fh:write("\n")

          if vhost.ssl then
            writeVHostSSL(fh, cfg, vhost)
          end

          -- strip off possible domain name (when using per-domain httpd.conf include files)
          dst = i:find('/#/')
          local include_conf
          if dst ~= nil and LC.fs.is_file(opts.path .. "/conf/" .. i:sub(dst+3) .. ".httpd.conf") then
            include_conf = opts.path .. "/conf/" .. i:sub(dst+3) .. ".httpd.conf"
            dst = i:sub(1, dst-1)
          else
            dst = i
          end

          if opts.path then
            writeVHostDocRoot(fh, cfg, opts, vhost, opts.path .. "/" .. LC.web.HTDOCS_PATH .. "/" .. dst, php)
          end

          -- enable RewriteEngine
          fh:write("    <IfModule mod_rewrite.c>\n")
          fh:write("        RewriteEngine On\n\n")

          -- /.well-known/ hook for domain validation (SSL certificates):
          fh:write("        RewriteMap  \"lc\" \"int:tolower\"\n")
          fh:write("        RewriteCond %{REQUEST_URI} ^/\\.well-known/(.*/)?([^/]*)$\n")
          fh:write("        RewriteCond ", opts.path .. "/" .. LC.web.HTDOCS_PATH, "/.well-known/%1%2 !-f\n")
          fh:write("        RewriteCond ", LC.web.getWebRoot(), "/.well-known/htdocs/%1${lc:%{HTTP_HOST}}.%2 -f\n")
          fh:write("        RewriteRule / ", LC.web.getWebRoot(), "/.well-known/htdocs/%1${lc:%{HTTP_HOST}}.%2 [T=text/plain,L]\n")

          -- Web statistics:
          if opts.webstats and opts.webstats.domain and opts.webstats.path and LC.liveconfig.inTable(opts.webstats.domain, sd[php][i]) ~= nil then
            writeVHostStats(fh, cfg, opts)
          end

          fh:write("    </IfModule>\n\n") -- mod_rewrite.c

          if dst == "" or dst == "/" then
            -- add support for nice default error pages (only for domains configured in webspace root directory!)
            fh:write([[
    <LocationMatch "^/$">
        ErrorDocument 403 /.errorFiles/coming-soon.html
        ErrorDocument 404 /.errorFiles/coming-soon.html
    </LocationMatch>

]])
          end

          -- write ACME challenges
          writeACME(fh, cfg, opts, vhost)

          -- include custom configuration
          writeInclude(fh, opts.path, include_conf)

          fh:write("</VirtualHost>\n")

        end -- for i in pairs(sd[php]) ...
      end -- for php in pairs(sd) ...
    end -- sd_count > 0

    -- Step 2: configure all proxy domains:
    if prx_count > 0 then
      for i in pairs(prx) do
        writeVHostIPs(fh, vhost)

        table.sort(prx[i])
        if string.match(prx[i][1], "^%*%.") then
          prx[i][#prx[i] + 1] = prx[i][1]
          prx[i][1] = "_catchall_." .. string.sub(prx[i][1], 3)
        end

        local dst
        for dst in pairs(prx[i]) do
          if dst == 1 then
            fh:write("    ServerName ", prx[i][dst], "\n")
            cnames[#cnames+1] = prx[i][dst]
          else
            fh:write("    ServerAlias ", prx[i][dst], "\n")
          end
        end
        fh:write("\n")

        if vhost.ssl then
          writeVHostSSL(fh, cfg, vhost)
        end

        -- /.well-known/ hook for domain validation (SSL certificates):
        fh:write("    <IfModule mod_rewrite.c>\n")
        fh:write("        RewriteEngine On\n")
        fh:write("        RewriteMap  \"lc\" \"int:tolower\"\n")
        fh:write("        RewriteCond %{REQUEST_URI} ^/\\.well-known/(.*/)?([^/]*)$\n")
        fh:write("        RewriteCond ", opts.path .. "/" .. LC.web.HTDOCS_PATH, "/.well-known/%1%2 !-f\n")
        fh:write("        RewriteCond ", LC.web.getWebRoot(), "/.well-known/htdocs/%1${lc:%{HTTP_HOST}}.%2 -f\n")
        fh:write("        RewriteRule / ", LC.web.getWebRoot(), "/.well-known/htdocs/%1${lc:%{HTTP_HOST}}.%2 [T=text/plain,L]\n")
        fh:write("    </IfModule>\n")

        -- write ACME challenges
        writeACME(fh, cfg, opts, vhost)

        -- strip off possible domain name (when using per-domain httpd.conf include files)
        dst = i:find('/#/')
        local include_conf
        if dst ~= nil and LC.fs.is_file(opts.path .. "/conf/" .. i:sub(dst+3) .. ".httpd.conf") then
          include_conf = opts.path .. "/conf/" .. i:sub(dst+3) .. ".httpd.conf"
          dst = i:sub(1, dst-1)
        else
          dst = i
        end

        fh:write("    <IfModule mod_proxy.c>\n")
        if string.sub(dst, -2) == '/*' then dst = string.sub(dst, 1, string.len(dst)-1) end
        fh:write("      ProxyPass        \"/\" \"", dst, "\"\n")
        fh:write("      ProxyPassReverse \"/\" \"", dst, "\"\n")
        if string.match(i, "^https://") then
          fh:write("\n")
          fh:write("      <IfModule mod_ssl.c>\n")
          fh:write("        SSLProxyEngine On\n")
          fh:write("      </IfModule>\n")
        end
        fh:write("    </IfModule>\n")
        fh:write("    <IfModule !mod_proxy.c>\n")
        fh:write("      <IfModule mod_rewrite.c>\n")
        fh:write("        RewriteEngine On\n")
        fh:write("        RewriteCond %{ENV:REDIRECT_STATUS} !=503\n")
        fh:write("        RewriteRule ^.* - [L,R=503]\n")
        fh:write("        ErrorDocument 503 /.errorFiles/no-proxy.html\n")
        fh:write("      </IfModule>\n")
        fh:write("    </IfModule>\n\n")

        -- include custom configuration
        writeInclude(fh, opts.path, include_conf)

        fh:write("</VirtualHost>\n")

      end -- for i in pairs(prx) ...
    end

    -- Step 3: configure all redirects:
    if r301_count > 0 or r302_count > 0 then
      -- create table "dl" containing all domains with 301/302 redirect
      -- so we can configure all of them together in one single <VirtualHost>
      local dl = { }
      local j
      for i in pairs(r301) do
        for j in pairs(r301[i]) do
          dl[#dl + 1] = r301[i][j]
        end
      end
      for i in pairs(r302) do
        for j in pairs(r302[i]) do
          dl[#dl + 1] = r302[i][j]
        end
      end

      writeVHostIPs(fh, vhost)

      table.sort(dl)
      if string.match(dl[1], "^%*%.") then
        dl[#dl + 1] = dl[1]
        dl[1] = "_catchall_." .. string.sub(dl[1], 3)
      end

      for i in pairs(dl) do
        if i == 1 then
          fh:write("    ServerName ", dl[i], "\n")
          cnames[#cnames+1] = dl[i]
        else
          fh:write("    ServerAlias ", dl[i], "\n")
        end
      end
      fh:write("\n")

      if vhost.ssl then
        writeVHostSSL(fh, cfg, vhost)
      end

      -- write ACME challenges (BEFORE redirecting!)
      writeACME(fh, cfg, opts, vhost)

      -- enable RewriteEngine
      fh:write("    <IfModule mod_rewrite.c>\n")
      fh:write("        RewriteEngine On\n\n")

      -- /.well-known/ hook for domain validation (SSL certificates):
      fh:write("        RewriteMap  \"lc\" \"int:tolower\"\n")
      fh:write("        RewriteCond %{REQUEST_URI} ^/\\.well-known/(.*/)?([^/]*)$\n")
      fh:write("        RewriteCond ", opts.path .. "/" .. LC.web.HTDOCS_PATH, "/.well-known/%1%2 !-f\n")
      fh:write("        RewriteCond ", LC.web.getWebRoot(), "/.well-known/htdocs/%1${lc:%{HTTP_HOST}}.%2 -f\n")
      fh:write("        RewriteRule / ", LC.web.getWebRoot(), "/.well-known/htdocs/%1${lc:%{HTTP_HOST}}.%2 [T=text/plain,L]\n")

      -- Web statistics:
      if opts.webstats and opts.webstats.domain and opts.webstats.path and LC.liveconfig.inTable(opts.webstats.domain, dl) ~= nil then
        writeVHostStats(fh, cfg, opts)
      end

      writeVHostRedirects(fh, r301, r302)

      fh:write("    </IfModule>\n\n") -- mod_rewrite.c

      -- include custom configuration
      writeInclude(fh, opts.path, nil)

      fh:write("</VirtualHost>\n")
    end -- r301_count > 0 or r302_count > 0

    -- Step 4: configure all domains pointing to a web application:
    if app_count > 0 then
      for php in pairs(app) do
        for i in pairs(app[php]) do

          writeVHostIPs(fh, vhost)

          table.sort(app[php][i])
          if string.match(app[php][i][1], "^%*%.") then
            app[php][i][#app[php][i] + 1] = app[php][i][1]
            app[php][i][1] = "_catchall_." .. string.sub(app[php][i][1], 3)
          end

          local d
          for d in pairs(app[php][i]) do
            if d == 1 then
              fh:write("    ServerName ", app[php][i][d], "\n")
              cnames[#cnames+1] = app[php][i][d]
            else
              fh:write("    ServerAlias ", app[php][i][d], "\n")
            end
          end
          fh:write("\n")

          if vhost.ssl then
            writeVHostSSL(fh, cfg, vhost)
          end

          -- strip off possible domain name (when using per-domain httpd.conf include files)
          dst = i:find('/#/')
          local include_conf
          if dst ~= nil and LC.fs.is_file(opts.path .. "/conf/" .. i:sub(dst+3) .. ".httpd.conf") then
            include_conf = opts.path .. "/conf/" .. i:sub(dst+3) .. ".httpd.conf"
            dst = i:sub(1, dst-1)
          else
            dst = i
          end

          if opts.path then
            writeVHostDocRoot(fh, cfg, opts, vhost, opts.path .. "/apps/" .. dst, php)
          end

          -- enable RewriteEngine
          fh:write("    <IfModule mod_rewrite.c>\n")
          fh:write("        RewriteEngine On\n\n")

          -- /.well-known/ hook for domain validation (SSL certificates):
          fh:write("        RewriteMap  \"lc\" \"int:tolower\"\n")
          fh:write("        RewriteCond %{REQUEST_URI} ^/\\.well-known/(.*/)?([^/]*)$\n")
          fh:write("        RewriteCond ", opts.path .. "/" .. LC.web.HTDOCS_PATH, "/.well-known/%1%2 !-f\n")
          fh:write("        RewriteCond ", LC.web.getWebRoot(), "/.well-known/htdocs/%1${lc:%{HTTP_HOST}}.%2 -f\n")
          fh:write("        RewriteRule / ", LC.web.getWebRoot(), "/.well-known/htdocs/%1${lc:%{HTTP_HOST}}.%2 [T=text/plain,L]\n")

          -- Web statistics:
          if opts.webstats and opts.webstats.domain and opts.webstats.path and LC.liveconfig.inTable(opts.webstats.domain, app[php][i]) ~= nil then
            writeVHostStats(fh, cfg, opts)
          end

          fh:write("    </IfModule>\n\n") -- mod_rewrite.c

          -- write ACME challenges
          writeACME(fh, cfg, opts, vhost)

          -- include custom configuration
          writeInclude(fh, opts.path, include_conf)

          fh:write("</VirtualHost>\n")

        end -- for i in pairs(app[php]) ...
      end -- for php in pairs(app) ...
    end -- app_count > 0

    -- Step 5: configure all website builder domains
    if bldr_count > 0 then
      for i in pairs(bldr) do
        writeVHostIPs(fh, vhost)

        table.sort(bldr[i])
        if string.match(bldr[i][1], "^%*%.") then
          bldr[i][#bldr[i] + 1] = bldr[i][1]
          bldr[i][1] = "_catchall_." .. string.sub(bldr[i][1], 3)
        end

        local dst
        for dst in pairs(bldr[i]) do
          if dst == 1 then
            fh:write("    ServerName ", bldr[i][dst], "\n")
            cnames[#cnames+1] = bldr[i][dst]
          else
            fh:write("    ServerAlias ", bldr[i][dst], "\n")
          end
        end
        fh:write("\n")

        if vhost.ssl then
          writeVHostSSL(fh, cfg, vhost)
        end

        -- strip off possible domain name (when using per-domain httpd.conf include files)
        dst = i:find('/#/')
        local include_conf
        if dst ~= nil and LC.fs.is_file(opts.path .. "/conf/" .. i:sub(dst+3) .. ".httpd.conf") then
          include_conf = opts.path .. "/conf/" .. i:sub(dst+3) .. ".httpd.conf"
          dst = i:sub(1, dst-1)
        else
          dst = i
        end

        if opts.path then
          -- in this case, 'dst' contains the path 'apps/<sitename>'
          -- use empty value ('') for php, so we use server default PHP version
          writeVHostDocRoot(fh, cfg, opts, vhost, opts.path .. "/" .. dst, '')
        end

        -- /.well-known/ hook for domain validation (SSL certificates):
        fh:write("    <IfModule mod_rewrite.c>\n")
        fh:write("        RewriteEngine On\n")
        fh:write("        RewriteMap  \"lc\" \"int:tolower\"\n")
        fh:write("        RewriteCond %{REQUEST_URI} ^/\\.well-known/(.*/)?([^/]*)$\n")
        fh:write("        RewriteCond ", opts.path .. "/" .. LC.web.HTDOCS_PATH, "/.well-known/%1%2 !-f\n")
        fh:write("        RewriteCond ", LC.web.getWebRoot(), "/.well-known/htdocs/%1${lc:%{HTTP_HOST}}.%2 -f\n")
        fh:write("        RewriteRule / ", LC.web.getWebRoot(), "/.well-known/htdocs/%1${lc:%{HTTP_HOST}}.%2 [T=text/plain,L]\n")
        fh:write("    </IfModule>\n")

        -- write ACME challenges
        writeACME(fh, cfg, opts, vhost)

        -- strip off possible domain name (when using per-domain httpd.conf include files)
        dst = i:find('/#/')
        local include_conf
        if dst ~= nil and LC.fs.is_file(opts.path .. "/conf/" .. i:sub(dst+3) .. ".httpd.conf") then
          include_conf = opts.path .. "/conf/" .. i:sub(dst+3) .. ".httpd.conf"
          dst = i:sub(1, dst-1)
        else
          dst = i
        end

        -- include custom configuration
        writeInclude(fh, opts.path, include_conf)

        fh:write("</VirtualHost>\n")

      end -- for i in pairs(prx) ...
    end

  else
    fh:write("# Unknown configuration mode '", CONFIGMODE, "'!")
  end

  if vhost.ssl then
    fh:write("</IfModule>\n")
  end

  -- return array with canonical server names
  return cnames
end

-- ---------------------------------------------------------------------------
-- configureVHost(cfg, opts)
--
-- Configure virtual host for Apache httpd
-- ---------------------------------------------------------------------------
function configureVHost(cfg, opts, cnames)

  LC.log.print(LC.log.DEBUG, "configureVHost(apache) called")

  local vhostpath  = cfg["available_dir"] or cfg["enabled_dir"]
  if opts and opts.prefix then vhostpath = opts.prefix .. vhostpath end -- use prefix (for testing etc.)
  local filename = vhostpath .. "/" .. opts["name"] .. ".conf"
  local tempfile = filename .. ".tmp"
--  tempfile = "/tmp/test.conf"
  local i = 0
  local key
  local vhosts = 0    -- number of <VirtualHost> sections

  if LC.fs.is_file(filename .. ".lock") then
    -- don't modify configuration
    LC.log.print(LC.log.INFO, "Not modifying ", filename, ": configuration locked")
    return true
  end

  -- create temporary config file
  local fh, msg = io.open(tempfile, "w")
  if fh == nil then
    LC.log.print(LC.log.ERR, "Can't open '", tempfile, "' for writing: ", msg)
    return false, "Can't open '" .. tempfile .. "' for writing: " .. msg
  end
  LC.liveconfig.writeHeader(fh)
  fh:write("# Automatically created by LiveConfig - do not modify!\n")
  fh:write("# Created at: ", os.date("%Y-%m-%d %H:%M:%S %Z"), "\n")
  fh:write("# ----------------------------------------------------------------------------\n\n")

  -- now check every single vhost, configure it if it's for this server:
  local vhostIdx
  if type(opts.vhosts) == "table" then
    -- create all VirtualHost sections for regular (non-wildcard) domains:
    for vhostIdx=1, #opts.vhosts do
      while opts.vhosts[vhostIdx].server == cfg.type do
        local vhost = opts.vhosts[vhostIdx]

        fh:write("# IP group: '", vhost.ipgroup, "'\n")
        if vhost.ips == nil then
          fh:write("# -> no IPs configured!\n")
          break
        end

        local dn = { }  -- list with all regular domain names
        for key in pairs(vhost.domains) do
          if not string.match(key, "^%*%.") then
            dn[#dn + 1] = key
          end
        end
        table.sort(dn)

        local cn
        -- configure all regular domains
        if #dn > 0 then
          cn = buildConfig(fh, cfg, opts, vhost, dn)
          for i, key in pairs(cn) do
            cnames[#cnames + 1] = key
            vhosts = vhosts + 1
          end
        end

        fh:write("\n")

        break -- leave 'while(true)' loop
      end -- while...
    end -- for...

    -- create all VirtualHost sections for wildcard domains:
    for vhostIdx=1, #opts.vhosts do
      while opts.vhosts[vhostIdx].server == cfg.type do
        local vhost = opts.vhosts[vhostIdx]

        fh:write("# IP group: '", vhost.ipgroup, "'\n")
        if vhost.ips == nil then
          fh:write("# -> no IPs configured!\n")
          break
        end

        local wc = { }  -- list with all wildcard domain names
        for key in pairs(vhost.domains) do
          if string.match(key, "^%*%.") then
            wc[#wc + 1] = key
          end
        end
        table.sort(wc)

        local cn
        -- configure all wildcard domains
        if #wc > 0 then
          cn = buildConfig(fh, cfg, opts, vhost, wc)
          for i, key in pairs(cn) do
            cnames[#cnames + 1] = key
            vhosts = vhosts + 1
          end
        end

        fh:write("\n")

        break -- leave 'while(true)' loop
      end -- while...
    end -- for...

  end -- if type(opts.vhosts) == "table"

  -- close (temporary) config file
  LC.liveconfig.writeFooter(fh)
  fh:close()

  -- rename config file (atomic)
  LC.fs.rename(tempfile, filename)

  if vhosts == 0 then
    -- disable configuration, remove config file
    if cfg["disable_cmd"] ~= nil then
      -- use "os.execute" instead of "LC.exec" to suppress output
      os.execute(cfg["disable_cmd"] .. " " .. opts["name"] .. ".conf")
    end
    os.remove(filename)
    LC.log.print(LC.log.INFO, "Removing '", filename, "' (no vHosts configured)")
  else
    -- if not done yet, enable config
    if cfg["enable_cmd"] ~= nil then
      -- use "os.execute" instead of "LC.exec" to suppress output
      os.execute(cfg["enable_cmd"] .. " " .. opts["name"] .. ".conf")
    end
  end

  if opts.hasPHP == 2 then
    -- create/update FastCGI wrapper scripts
    local phpidx
    for phpidx in pairs(LCS.web.phpVersions) do
      local php = LCS.web.phpVersions[phpidx]
      local inputfile = php.ini
      local code = php.code
      os.execute("chattr -i " .. opts.path .. "/conf/" .. code)
      local pfh, msg = io.open(opts.path .. "/conf/" .. code .. "/php-fcgi-starter.tmp", "w")
      if pfh == nil then
        os.execute("chattr +i " .. opts.path .. "/conf/" .. code)
        LC.log.print(LC.log.ERR, "Can't open '", opts.path .. "/conf/" .. code .. "/php-fcgi-starter.tmp", "' for writing: ", msg)
      else
        pfh:write("#!/bin/sh\n")
        pfh:write("umask 0022\n")
        if LC.fs.is_file(opts.path .. "/.php5/php.ini") then
          -- use custom php.ini file
          pfh:write("PHPRC=", opts.path, "/.php5/\n")
        elseif LC.fs.is_file(opts.path .. "/conf/" .. code .. "/php.ini") then
          -- use php.ini file customized by LiveConfig
          pfh:write("PHPRC=", opts.path, "/conf/" .. code .. "/\n")
        end
        pfh:write("export PHPRC\n")
        -- set PHP_FCGI_MAX_REQUESTS
        pfh:write("PHP_FCGI_MAX_REQUESTS=5000\n")
        pfh:write("export PHP_FCGI_MAX_REQUESTS\n")
        -- export TMP and TMPDIR variable for custom temp directory:
        pfh:write("TMP=", opts.path, "/tmp\n")
        pfh:write("export TMP\n")
        pfh:write("TMPDIR=", opts.path, "/tmp\n")
        pfh:write("export TMPDIR\n")
        pfh:write("exec " .. php.bin .. "\n")
        pfh:close()
        if LC.fs.is_file(opts.path .. "/conf/" .. code .. "/php-fcgi-starter") then
          -- remove "immutable" flag:
          os.execute("chattr -i " .. opts.path .. "/conf/" .. code .. "/php-fcgi-starter")
        end
        LC.fs.setperm(opts.path .. "/conf/" .. code .. "/php-fcgi-starter.tmp", 550, opts.user, opts.group)
        LC.fs.rename(opts.path .. "/conf/" .. code .. "/php-fcgi-starter.tmp", opts.path .. "/conf/" .. code .. "/php-fcgi-starter")
        -- set SELinux permissions
        if LC.distribution.family == "RedHat" then
          os.execute("chcon --type=httpd_sys_script_exec_t \"" .. opts.path .. "/conf/" .. code .. "/php-fcgi-starter\"")
        end
        -- set "immutable" flag:
        os.execute("chattr +i " .. opts.path .. "/conf/" .. code .. "/php-fcgi-starter")
        os.execute("chattr +i " .. opts.path .. "/conf/" .. code)
      end
    end
  end

  -- reload web server configuration
  LC.timeout.set('apache.reload', RELOAD_MIN, RELOAD_MAX)

  return true

end

-- ---------------------------------------------------------------------------
-- reload()
--
-- Reload Apache configuration
-- ---------------------------------------------------------------------------
function reload()
  LC.log.print(LC.log.DEBUG, "apache.reload() called")
  -- get configuration
  local cfg = LC.web.getConfig('apache')
  if cfg == nil then
    LC.log.print(LC.log.ERR, "apache.reload(): no configuration for 'apache' available!?")
    return
  end
  -- get reload command
  if cfg.reload_cmd == nil then
    LC.log.print(LC.log.ERR, "apache.reload(): no reload command for 'apache' available!?")
    return
  end

  -- reload service
  LC.exec(cfg.reload_cmd)

end

-- ---------------------------------------------------------------------------
-- restart()
--
-- Restart Apache web server
-- ---------------------------------------------------------------------------
function restart()
  LC.log.print(LC.log.DEBUG, "apache.restart() called")
  -- get configuration
  local cfg = LC.web.getConfig('apache')
  if cfg == nil then
    LC.log.print(LC.log.ERR, "apache.restart(): no configuration for 'apache' available!?")
    return
  end
  -- get restart command
  if cfg.restart_cmd == nil then
    LC.log.print(LC.log.ERR, "apache.restart(): no restart command for 'apache' available!?")
    return
  end

  -- restart service
  LC.exec(cfg.restart_cmd)

end

-- ---------------------------------------------------------------------------
-- deleteAccount()
--
-- Delete account/vHost
-- ---------------------------------------------------------------------------
function deleteAccount(cfg, name)
  local vhostpath  = cfg["available_dir"] or cfg["enabled_dir"]
  local filename = vhostpath .. "/" .. name .. ".conf"

  LC.log.print(LC.log.DEBUG, "apache.deleteAccount(", name, ") (config file: '", filename, "'")

  if not LC.fs.is_file(filename) then return end

  if cfg["disable_cmd"] ~= nil then
    LC.exec(cfg["disable_cmd"] .. " " .. name .. ".conf")
  end
  os.remove(filename)

  -- reload web server configuration
  LC.timeout.set('apache.reload', RELOAD_MIN, RELOAD_MAX)

end

-- <EOF>----------------------------------------------------------------------
