--  _    _          ___           __ _     (R)
-- | |  (_)_ _____ / __|___ _ _  / _(_)__ _
-- | |__| \ V / -_) (__/ _ \ ' \|  _| / _` |
-- |____|_|\_/\___|\___\___/_||_|_| |_\__, |
--                                    |___/
-- Copyright (c) 2009-2019 LiveConfig GmbH.
-- ---------------------------------------------------------------------------
-- common/lua/backup.lua
-- Backup service management
-- $Id$
-- ---------------------------------------------------------------------------

-- Imported functions
local LC = LC
local LCS = LCS
local io = io
local os = os
local string = string

-- Module declaration
module("backup")

-- Exported functions
--   detect()
--   install()
--   uninstall()

-- ---------------------------------------------------------------------------
-- getBorgBinaryVersion()
--
-- Return the binary version number from mysql
-- ---------------------------------------------------------------------------
local function getBorgBinaryVersion()
  local handle = io.popen("/usr/bin/borg -V", "r")
  local ret = handle:read("*a")
  handle:close()
  local v = string.match(ret, "borg (%d+%.%d+[%.%d+]*)")
  return v
end

-- ---------------------------------------------------------------------------
-- getResticBinaryVersion()
--
-- Return the binary version number from mysql
-- ---------------------------------------------------------------------------
local function getResticBinaryVersion()
  local handle = io.popen("/usr/bin/restic version", "r")
  local ret = handle:read("*a")
  handle:close()
  local v = string.match(ret, "restic (%d+%.%d+[%.%d+]*)")
  return v
end

-- ---------------------------------------------------------------------------
-- getMysqlDumpBinaryVersion()
--
-- Return the binary version number from mysql
-- ---------------------------------------------------------------------------
local function isMysqlDumpInstalled()
  local handle = io.popen("/usr/bin/mysqldump -V", "r")
  if handle == nil then
    return false
  end
  handle:close()
  return true
end

-- ---------------------------------------------------------------------------
-- detect()
--
-- Detect if the following binaries are installed:
-- borg
-- restic
-- mysqldump

-- all relevant informations:
--   borg               => true when exists
--   borg_version       => binary version of borg
--   borg_pkg_version   => package version number (eg. "2.2.3-4+etch11")
--   restic             => true when exists
--   restic_version     => binary version of restic
--   restic_pkg_version => package version number (eg. "2.2.3-4+etch11")
-- ---------------------------------------------------------------------------
function detect()
  local pkg, v
  local data = {}

  -- check if DB server software was already detected; avoid double-detection
  -- also across multiple threads
  LC.mutex.lock("backup.detect")

  if LCS.backup ~= nil then
    if LCS.backup.data ~= nil then
      -- DB server software already detected; load values from global LCS storage
      LC.mutex.unlock("backup.detect")
      return LCS.backup.data
    end
  end

  local backup = {
    ["type"] = "backup",
    ["borg"] = false,
    ["restic"] = false
  }

  -- first check for distribution-specific packages
  pkg, v = LC.distribution.hasPackage('borgbackup')
  if pkg ~= nil then
    LC.log.print(LC.log.DEBUG, "Found package '", pkg, "' (Version ", v, ")")
    backup.borg = true
    backup.borg_version = getBorgBinaryVersion()
    backup.borg_pkg_version = v
  end
  pkg, v = LC.distribution.hasPackage('restic')
  if pkg ~= nil then
    LC.log.print(LC.log.DEBUG, "Found package '", pkg, "' (Version ", v, ")")
    backup.restic = true
    backup.restic_version = getResticBinaryVersion()
    backup.restic_pkg_version = v
  end

  backup.mysqldump = isMysqlDumpInstalled()
  data[#data + 1] = backup

  LCS.backup = {
    generated = true,
    data      = data    -- LCS.backup.data = (local)data
  }

  LC.mutex.unlock("backup.detect")

  return data
end

-- ---------------------------------------------------------------------------
-- install()
--
-- Install & enable the backup service (lcbackup)
-- ---------------------------------------------------------------------------
function install()

  LC.log.print(LC.log.INFO, "Installing backup service (lcbackup)")

  -- acquire mutex
  LC.mutex.lock("backup.install")

  local ret, stdout, stderr

  -- install service (systemd)
  if LC.fs.is_dir("/lib/systemd/system") and not LC.fs.is_file("/lib/systemd/system/lcbackup.service") then
    LC.exec("cp --preserve=all " .. LC.liveconfig.libexecdir .. "/lcbackup.service /lib/systemd/system/")
    if LC.fs.is_file("/bin/systemctl") then
      ret, stdout, stderr = LC.exec("/bin/systemctl enable lcbackup")
      if ret ~= 0 then
        LC.mutex.unlock("backup.install")
        return false, stderr
      end
    end
  end

  -- lcbackup: install init script
  if not LC.fs.is_file("/etc/init.d/lcbackup") then
    LC.exec("cp " .. LC.liveconfig.libexecdir .. "/lcbackup.init /etc/init.d/lcbackup")
    LC.fs.setperm("/etc/init.d/lcbackup", "0700", "root", "root")

    if LC.fs.is_file("/usr/sbin/update-rc.d") then
      ret, stdout, stderr = LC.exec("/usr/sbin/update-rc.d lcbackup defaults")
    elseif LC.fs.is_file("/sbin/chkconfig") then
      ret, stdout, stderr = LC.exec("/sbin/chkconfig --add lcbackup")
    elseif LC.fs.is_file("/sbin/rc-update") then
      ret, stdout, stderr = LC.exec("/sbin/rc-update add lcbackup default")
    else
      ret = 1
      stderr = "Can't install init script 'lcbackup' - no supported tool found"
    end

    if ret ~= 0 then
      LC.mutex.unlock("backup.install")
      return false, stderr
    end

  end

  -- start lcbackup
  if LC.fs.is_file("/sbin/service") then
    ret, stdout, stderr = LC.exec("/sbin/service lcbackup start")
  else
    ret, stdout, stderr = LC.exec("/etc/init.d/lcbackup start")
  end

  if ret ~= 0 then
    LC.mutex.unlock("backup.install")
    return false, stderr
  end

  -- release mutex
  LC.mutex.unlock("backup.install")

  return true

end

-- ---------------------------------------------------------------------------
-- uninstall()
--
-- Uninstall backup service (lcbackup)
-- ---------------------------------------------------------------------------
function uninstall()

  LC.log.print(LC.log.INFO, "Uninstalling backup service (lcbackup)")

  -- acquire mutex
  LC.mutex.lock("backup.install")

  -- quit lcbackup
  if LC.fs.is_file("/sbin/service") then
    LC.exec("/sbin/service lcbackup stop")
  else
    LC.exec("/etc/init.d/lcbackup stop")
  end

  -- disable lcbackup
  if LC.fs.is_file("/lib/systemd/system/lcbackup") then
    -- disable service (systemd)
    if LC.fs.is_file("/bin/systemctl") then
      LC.exec("/bin/systemctl disable lcbackup")
    end
    -- remove service file
    os.remove("/lib/systemd/system/lcbackup")
  end

  if LC.fs.is_file("/etc/init.d/lcbackup") then
    if LC.fs.is_file("/usr/sbin/update-rc.d") then
      LC.exec("/usr/sbin/update-rc.d -f lcbackup remove")
    elseif LC.fs.is_file("/sbin/chkconfig") then
      LC.exec("/sbin/chkconfig --del lcbackup")
    elseif LC.fs.is_file("/sbin/rc-update") then
      LC.exec("/sbin/rc-update del lcbackup default")
    else
      LC.log.print(LC.log.ERR, "Can't uninstall init script 'lcbackup' - no supported tool found")
    end
    os.remove("/etc/init.d/lcbackup")
  end

  -- release mutex
  LC.mutex.unlock("backup.install")

  return true

end

-- <EOF>----------------------------------------------------------------------
