--  _    _          ___           __ _     (R)
-- | |  (_)_ _____ / __|___ _ _  / _(_)__ _
-- | |__| \ V / -_) (__/ _ \ ' \|  _| / _` |
-- |____|_|\_/\___|\___\___/_||_|_| |_\__, |
--                                    |___/
-- Copyright (c) 2009-2019 LiveConfig GmbH.
-- ---------------------------------------------------------------------------
-- common/lua/cron.lua
-- Cron tab management
-- $Id$
-- ---------------------------------------------------------------------------

-- Imported functions
local LC = LC
local io = io
local os = os
local type = type
local ipairs = ipairs

-- Module declaration
module("cron")

-- Exported functions
--   update()

-- ---------------------------------------------------------------------------
-- update()
--
-- Update the cron table for a user
-- ---------------------------------------------------------------------------
function update(user, mailto, crontab, suspended)

  LC.log.print(LC.log.INFO, "Updating crontab for '", user, "'")

  local fh, msg = LC.popen("/usr/bin/crontab -u " .. user .. " -")
  if fh == nil then
    LC.log.print(LC.log.ERR, "Can't execute '/usr/bin/crontab -u ", user, " -': ", msg)
    return false, "Can't execute '/usr/bin/crontab -u ", user, " -': " .. msg
  end

  -- now write new crontab into this process:
  fh:write("# Crontab for user ", user, "\n")
  fh:write("# Automatically created by LiveConfig - DO NOT EDIT\n\n")

  if suspended then
    fh:write("# subscription is suspended - all cron jobs disabled.\n")
  else

    if type(crontab) == "table" then
      local k, v
      for k, v in ipairs(crontab) do
        -- check for per cronjob email
        if v["email"] ~= nil then
          fh:write("\nMAILTO=\"", v["email"], "\"\n")
        else
          -- else check for global mailto
          if mailto == nil then
            fh:write("MAILTO=\"\"\n")
          else
            fh:write("MAILTO=\"", mailto, "\"\n")
          end
        end
        local line
        line = v["min"] .. " " .. v["hour"] .. " " .. v["day"] .. " " .. v["month"] .. " " .. v["dow"] .. "\t" .. v["cmd"]
        fh:write(line, "\n")
      end
    end

  end

  fh:write("\n# <EOF>\n")
  -- close pipe:
  fh:close()

  -- read any output from STDERR
  while fh ~= nil do
    local line = fh:readErr()
    if line == nil then break end
    LC.log.print(LC.log.ERR, "CRON[" .. user .. "]: '" .. line .. "'")
  end

  local st = fh:wait()
  if st ~= nil and st ~= 0 then
    LC.log.print(LC.log.ERR, "crontab(" .. user .. ") returned with exit code " .. st)
    return false, "crontab(" .. user .. ") returned with exit code " .. st
  end

  return true

end

-- ---------------------------------------------------------------------------
-- delete()
--
-- Delete cron tab for a user
-- ---------------------------------------------------------------------------
function delete(user)

  local rc = LC.exec("/usr/bin/crontab -u " .. user .. " -r")

end

-- ---------------------------------------------------------------------------

-- register hooks
if LC.hooks then
  LC.hooks.add("LC.users.delUser", delete)
end

-- <EOF>----------------------------------------------------------------------
