--  _    _          ___           __ _     (R)
-- | |  (_)_ _____ / __|___ _ _  / _(_)__ _
-- | |__| \ V / -_) (__/ _ \ ' \|  _| / _` |
-- |____|_|\_/\___|\___\___/_||_|_| |_\__, |
--                                    |___/
-- Copyright (c) 2009-2019 LiveConfig GmbH.
-- ---------------------------------------------------------------------------
-- common/lua/dkim.lua
-- DKIM configuration/management
-- $Id$
-- ---------------------------------------------------------------------------

-- Imported functions
local LC = LC
local LCS = LCS
local os = os
local io = io
local ipairs = ipairs
local string = string

-- Module declaration
module("dkim")

-- Exported functions
--   detect()
--   install()
--   uninstall()
--   addDomain()
--   deleteDomain()
--   restart()

-- ---------------------------------------------------------------------------
-- detect()
--
-- Detect installed DKIM package
-- ---------------------------------------------------------------------------
function detect()
  local data

  -- check if smtp server software was already detected; avoid double-detection
  -- also across multiple threads
  LC.mutex.lock("dkim.detect")

  if LCS.dkim ~= nil and LCS.dkim.data ~= nil then
    -- DKIM software already detected; load values from global LCS storage
    LC.mutex.unlock("dkim.detect")
    return LCS.dkim.data
  end

  -- package is known as "opendkim" on all distributions
  local pkg, v = LC.distribution.hasPackage('opendkim')
  if pkg ~= nil then
    LC.log.print(LC.log.DEBUG, "Found package '", pkg, "' (Version ", v, ")")
    data = {
      pkgname       = pkg,
      pkgversion    = v,
      configured    = false,
      configpath    = "/etc",
      configfile    = "/etc/opendkim.conf",
      statusfile    = "/etc/opendkim/liveconfig.status",
      start_cmd     = "/etc/init.d/opendkim start",
      stop_cmd      = "/etc/init.d/opendkim stop",
      restart_cmd   = "/etc/init.d/opendkim restart",
      optionsfile   = "/etc/default/opendkim",
      keypath       = "/etc/opendkim/keys",
      keytable      = "/etc/opendkim/keytable",
      signingtable  = "/etc/opendkim/signingtable"
    }
    if LC.fs.is_file("/etc/opendkim/opendkim.conf") then
      -- OpenSUSE >=15
      data.configfile = "/etc/opendkim/opendkim.conf"
    end
    if LC.distribution.family == "RedHat" then
      -- "restart" doesn't work reliably under CentOS!
      data.restart_cmd = "/etc/init.d/opendkim stop; sleep 2; /etc/init.d/opendkim start"
    end
    if LC.fs.is_file("/bin/systemctl") or LC.fs.is_file("/usr/bin/systemctl") then
      data.start_cmd = "systemctl start opendkim.service"
      data.stop_cmd = "systemctl stop opendkim.service"
      data.restart_cmd = "systemctl restart opendkim.service"
    end
    -- check for status file
    if LC.fs.is_file(data.statusfile) then
      local r = LC.liveconfig.readStatus(data.statusfile)
      if r and r.revision then
        data.configured = true
      end
    end
    LCS.dkim = { }
    LCS.dkim.data = data
  end

  LC.mutex.unlock("dkim.detect")
  
  return data
end

-- ---------------------------------------------------------------------------
-- install()
--
-- Install (configure) OpenDKIM
-- ---------------------------------------------------------------------------
function install()

  -- get configuration
  local cfg = detect()
  if cfg == nil then
    return false, "OpenDKIM not found"
  end

  if cfg.configured then
    -- already configured...
    return true
  end

  -- acquire mutex
  LC.mutex.lock("dkim.install")

  -- stop OpenDKIM (if running)
  -- (must be done *NOW*, because location of some files might change)
  LC.exec(cfg.stop_cmd)

  -- create opendkim directory if it doesn't exist (eg. on OpenSUSE)
  if not LC.fs.is_dir("/etc/opendkim") then
    LC.log.print(LC.log.INFO, "Creating directory '/etc/opendkim'")
    LC.fs.mkdir("/etc/opendkim")
  end

  -- create system group "opendkim" if it doesn't exist (eg. on OpenSUSE)
  if not LC.sys.group_exists("opendkim") then
    LC.log.print(LC.log.INFO, "Creating system group 'opendkim'")
    LC.exec("/usr/sbin/groupadd -r opendkim")
  end

  -- create system user "opendkim" if it doesn't exist (eg. on OpenSUSE)
  if not LC.sys.user_exists("opendkim") then
    LC.log.print(LC.log.INFO, "Creating system user 'opendkim'")
    LC.exec("useradd -r -d /etc/opendkim -g opendkim -s /bin/false opendkim")
  end

  -- add user "postfix" to group "opendkim", to allow writing to its socket:
  LC.log.print(LC.log.INFO, "Adding user 'postfix' to group 'opendkim'")
  LC.exec("usermod -a -G opendkim postfix")

  -- adjust owner of opendkim directory
  LC.fs.setperm("/etc/opendkim", 750, "root", "opendkim")

  -- create opendkim key directory if it doesn't exist
  if not LC.fs.is_dir(cfg.keypath) then
    LC.log.print(LC.log.INFO, "Creating directory '", cfg.keypath, "'")
    LC.fs.mkdir(cfg.keypath)
    -- adjust permissions
    LC.fs.setperm(cfg.keypath, 750, "opendkim", "root")
  end

  -- create opendkim pid directory if it doesn't exist
  if not LC.fs.is_dir("/var/run/opendkim") then
    LC.log.print(LC.log.INFO, "Creating directory '/var/run/opendkim'")
    LC.fs.mkdir("/var/run/opendkim")
    -- adjust permissions
    LC.fs.setperm(cfg.keypath, 750, "opendkim", "opendkim")
  end

  -- create opendkim socket directory if it doesn't exist
  if not LC.fs.is_dir("/var/spool/postfix/opendkim") then
    LC.log.print(LC.log.INFO, "Creating directory '/var/spool/postfix/opendkim'")
    LC.fs.mkdir("/var/spool/postfix/opendkim")
    -- adjust permissions
    LC.fs.setperm("/var/spool/postfix/opendkim", 750, "opendkim", "opendkim")
  end

  -- backup configuration file
  if LC.fs.is_file(cfg.configfile) and not LC.fs.is_file(cfg.configfile .. ".lcbak") then
    LC.fs.copy(cfg.configfile, cfg.configfile .. ".lcbak")
  end

  -- create new configuration file
  fh, msg = io.open(cfg.configfile .. ".tmp", "w")
  if fh == nil then
    LC.log.print(LC.log.ERR, "Can't open '", cfg.configfile, ".tmp' for writing: ", msg)
    return false, "Can't open '" .. cfg.configfile.. ".tmp' for writing: " .. msg
  end
  -- adjust owner & permissions - only OpenDKIM needs to read this file:
  LC.fs.setperm(cfg.configfile .. ".tmp", 640, "root", "opendkim")

  LC.liveconfig.writeHeader(fh)

  fh:write("# This is the configuration file for OpenDKIM\n")
  fh:write("# Created/Maintained by LiveConfig - DO NOT EDIT!\n")
  fh:write("# own options can be managed in '", cfg.configfile, ".local'\n\n")

  fh:write("AutoRestart             yes\n")
  fh:write("AutoRestartRate         2/1m\n")
  fh:write("Canonicalization        relaxed/simple\n")
  fh:write("KeyTable                ", cfg.keytable, "\n")
  fh:write("PidFile                 /var/run/opendkim/opendkim.pid\n")
  fh:write("SignatureAlgorithm      rsa-sha256\n")
  fh:write("SigningTable            refile:", cfg.signingtable, "\n")
  fh:write("Socket                  local:/var/spool/postfix/opendkim/opendkim.sock\n")
  fh:write("Syslog                  yes\n")
  fh:write("UMask                   007\n")
  fh:write("UserID                  opendkim\n\n")

  fh:write("Include                 ", cfg.configfile, ".local\n\n")
  LC.liveconfig.writeFooter(fh)
  fh:close()

  -- adjust /etc/default/opendkim
  if LC.fs.is_file('/etc/default/opendkim') then
    -- fix "SOCKET=" option on Ubuntu 16 systems
    -- (their systemd script is actually broken and doesn't accept value from opendkim.conf!)
    LC.exec("grep -q '^SOCKET=\"local:/var/run/opendkim/opendkim.sock\"' /etc/default/opendkim && sed -i -e 's/^SOCKET=/#SOCKET=/' /etc/default/opendkim && echo 'SOCKET=\"local:/var/spool/postfix/opendkim/opendkim.sock\"' >>/etc/default/opendkim")

    -- fix "RUNDIR=" option on Debian 9 systems
    -- (their systemd script is actually broken and doesn't accept value from opendkim.conf!)
    LC.exec("grep -q '^RUNDIR=/var/run/opendkim' /etc/default/opendkim && sed -i -e 's/^RUNDIR=\\(.*\\)/#RUNDIR=\\1\\nRUNDIR=\\/var\\/spool\\/postfix\\/opendkim/' /etc/default/opendkim")
    if LC.fs.is_file("/lib/opendkim/opendkim.service.generate") then
      LC.exec("/lib/opendkim/opendkim.service.generate")
    end
  end

  -- enable configuration
  LC.fs.rename(cfg.configfile .. ".tmp", cfg.configfile)

  -- create configuration file for local modifications
  if not LC.fs.is_file(cfg.configfile .. ".local") then
    fh, msg = io.open(cfg.configfile .. ".local", "w")
    if fh == nil then
      LC.log.print(LC.log.ERR, "Can't open '", cfg.configfile, ".local' for writing: ", msg)
      return false, "Can't open '" .. cfg.configfile.. ".local' for writing: " .. msg
    end
    -- adjust owner & permissions - only OpenDKIM needs to read this file:
    LC.fs.setperm(cfg.configfile .. ".local", 640, "root", "opendkim")
    -- this file need to be either completely empty, or contain at least one (active) configuration option.
    -- we keep it empty for now.
    fh:close()
  end

  -- create KeyTable (if not existing)
  if not LC.fs.is_file(cfg.keytable) then
    fh, msg = io.open(cfg.keytable, "w")
    if fh == nil then
      LC.log.print(LC.log.ERR, "Can't create '", cfg.keytable, "': ", msg)
      return false, "Can't create '" .. cfg.keytable.. "': " .. msg
    end
    -- adjust owner & permissions - only OpenDKIM needs to read this file:
    LC.fs.setperm(cfg.keytable, 640, "opendkim", "root")
    fh:close()
  end

  -- create SigningTable (if not existing)
  if not LC.fs.is_file(cfg.signingtable) then
    fh, msg = io.open(cfg.signingtable, "w")
    if fh == nil then
      LC.log.print(LC.log.ERR, "Can't create '", cfg.signingtable, "': ", msg)
      return false, "Can't create '" .. cfg.signingtable.. "': " .. msg
    end
    -- adjust owner & permissions - only OpenDKIM needs to read this file:
    LC.fs.setperm(cfg.signingtable, 640, "opendkim", "root")
    fh:close()
  end

  -- start OpenDKIM
  LC.log.print(LC.log.INFO, "Starting OpenDKIM...")
  LC.exec(cfg.start_cmd)

  -- write status file
  LC.liveconfig.writeStatus(cfg.statusfile, "dkim", 0, os.date("%Y-%m-%d %H:%M:%S %Z"))

  -- set "configured" flag
  cfg.configured = true

  -- release mutex
  LC.mutex.unlock("dkim.install")

  return true

end

-- ---------------------------------------------------------------------------
-- uninstall()
--
-- Uninstall/disable OpenDKIM
-- ---------------------------------------------------------------------------
function uninstall()

  -- get configuration
  local cfg = detect()
  if cfg == nil then
    return false, "OpenDKIM not found"
  end

  -- acquire mutex
  LC.mutex.lock("dkim.install")

  -- remove status file
  if LC.fs.is_file(cfg.statusfile) then
    os.remove(cfg.statusfile)
  end

  -- quit OpenDKIM
  LC.log.print(LC.log.INFO, "Stopping OpenDKIM...")
  LC.exec(cfg.stop_cmd)

  -- restore OpenDKIM configuration
  if LC.fs.is_file(cfg.configfile .. ".lcbak") then
    LC.fs.rename(cfg.configfile .. ".lcbak", cfg.configfile)
  end

  -- release mutex
  LC.mutex.unlock("dkim.install")

end

-- ---------------------------------------------------------------------------
local function filterLine(pattern, fhr, fhw)
  -- search/remove existing entry
  local line
  local found = false
  while true do
    line = fhr:read()
    if line == nil then break end
    if string.find(line, pattern) ~= nil then
      found = true
    else
      fhw:write(line, "\n")
    end
  end
end

-- ---------------------------------------------------------------------------
function addDomain(domain, key, selector)
  local cfg = LCS.dkim.data
  local keyfile = cfg.keypath .. "/" .. domain .. ".pem"
  local fhr, fhw, msg
  if selector == nil then selector = "mail" end

  -- acquire mutex
  LC.mutex.lock("dkim.install")

  -- open KeyTable, search for domain and add/update
  local pattern = '^[^.]+._domainkey.' .. domain .. ' '
  pattern = string.gsub(pattern, "%.", "%%.")
  pattern = string.gsub(pattern, "%-", "%%-")

  fhr, msg = io.open(cfg.keytable, "r")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", cfg.keytable, "' for reading: ", msg)
    LC.mutex.unlock("dkim.install")
    return false
  end

  fhw, msg = io.open(cfg.keytable .. ".tmp", "w")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", cfg.keytable, ".tmp' for writing: ", msg)
    LC.mutex.unlock("dkim.install")
    fhr:close()
    return false
  end

  LC.fs.setperm(cfg.keytable .. ".tmp", 640, "opendkim", "root")

  filterLine(pattern, fhr, fhw)
  fhr:close()
  fhw:write(selector, "._domainkey.", domain, " ", domain, ":", selector, ":", keyfile, "\n")
  fhw:close()
  LC.fs.rename(cfg.keytable .. ".tmp", cfg.keytable)

  -- open SigningTable, search for domain and add/update
  pattern = '^*@' .. domain .. ' [^.]+._domainkey.' .. domain
  pattern = string.gsub(pattern, "%.", "%%.")
  pattern = string.gsub(pattern, "%-", "%%-")

  fhr, msg = io.open(cfg.signingtable, "r")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", cfg.signingtable, "' for reading: ", msg)
    LC.mutex.unlock("dkim.install")
    return false
  end

  fhw, msg = io.open(cfg.signingtable .. ".tmp", "w")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", cfg.signingtable, ".tmp' for writing: ", msg)
    LC.mutex.unlock("dkim.install")
    fhr:close()
    return false
  end

  LC.fs.setperm(cfg.signingtable .. ".tmp", 640, "opendkim", "root")

  filterLine(pattern, fhr, fhw)
  fhr:close()
  fhw:write("*@", domain, " ", selector, "._domainkey.", domain, "\n")
  fhw:close()
  LC.fs.rename(cfg.signingtable .. ".tmp", cfg.signingtable)

  -- create/update key file
  fhw, msg = io.open(keyfile .. ".tmp", "w")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", keyfile, ".tmp' for writing: ", msg)
    LC.mutex.unlock("dkim.install")
    return false
  end
  LC.fs.setperm(keyfile .. ".tmp", 600, "opendkim", "root")
  fhw:write(key)
  fhw:close()
  LC.fs.rename(keyfile .. ".tmp", keyfile)

  -- release mutex
  LC.mutex.unlock("dkim.install")

  -- restart OpenDKIM service (not just reload, because UNIX socket to Postfix is broken)
  LC.timeout.set('dkim.restart', 10, 60)

  return true
end

-- ---------------------------------------------------------------------------
function deleteDomain(domain)
  if LCS.dkim == nil then return true end

  local cfg = LCS.dkim.data
  local keyfile = cfg.keypath .. "/" .. domain .. ".pem"
  local fhr, fhw, msg

  -- acquire mutex
  LC.mutex.lock("dkim.install")

  -- open KeyTable, search for domain and remove entry
  local pattern = '^[^.]+._domainkey.' .. domain .. ' '
  pattern = string.gsub(pattern, "%.", "%%.")
  pattern = string.gsub(pattern, "%-", "%%-")

  fhr, msg = io.open(cfg.keytable, "r")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", cfg.keytable, "' for reading: ", msg)
    LC.mutex.unlock("dkim.install")
    return false
  end

  fhw, msg = io.open(cfg.keytable .. ".tmp", "w")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", cfg.keytable, ".tmp' for writing: ", msg)
    LC.mutex.unlock("dkim.install")
    fhr:close()
    return false
  end

  LC.fs.setperm(cfg.keytable .. ".tmp", 640, "opendkim", "root")

  filterLine(pattern, fhr, fhw)
  fhr:close()
  fhw:close()
  LC.fs.rename(cfg.keytable .. ".tmp", cfg.keytable)

  -- open SigningTable, search for domain and remove entry
  pattern = '^*@' .. domain .. ' [^.]+._domainkey.' .. domain
  pattern = string.gsub(pattern, "%.", "%%.")
  pattern = string.gsub(pattern, "%-", "%%-")

  fhr, msg = io.open(cfg.signingtable, "r")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", cfg.signingtable, "' for reading: ", msg)
    LC.mutex.unlock("dkim.install")
    return false
  end

  fhw, msg = io.open(cfg.signingtable .. ".tmp", "w")
  if fhr == nil then
    LC.log.print(LC.log.ERR, "Can't open '", cfg.signingtable, ".tmp' for writing: ", msg)
    LC.mutex.unlock("dkim.install")
    fhr:close()
    return false
  end

  LC.fs.setperm(cfg.signingtable .. ".tmp", 640, "opendkim", "root")

  filterLine(pattern, fhr, fhw)
  fhr:close()
  fhw:close()
  LC.fs.rename(cfg.signingtable .. ".tmp", cfg.signingtable)

  -- remove key file
  os.remove(keyfile)

  -- release mutex
  LC.mutex.unlock("dkim.install")

  -- restart OpenDKIM service (not just reload, because UNIX socket to Postfix is broken)
  LC.timeout.set('dkim.restart', 10, 60)

  return true
end

-- ---------------------------------------------------------------------------
-- restart()
--
-- Restart OpenDKIM
-- ---------------------------------------------------------------------------
function restart()
  LC.log.print(LC.log.DEBUG, "dkim.restart() called")

  -- get configuration
  local cfg = LCS.dkim.data

  -- get restart command
  if cfg.restart_cmd == nil then
    LC.log.print(LC.log.ERR, "dkim.restart(): no restart command for 'dkim' available!?")
    return
  end

  -- restart service
  LC.exec(cfg.restart_cmd)
end

-- <EOF>----------------------------------------------------------------------
