<?php
/** _    _          ___           __ _     (R)
 * | |  (_)_ _____ / __|___ _ _  / _(_)__ _
 * | |__| \ V / -_) (__/ _ \ ' \|  _| / _` |
 * |____|_|\_/\___|\___\___/_||_|_| |_\__, |
 *                                    |___/
 * LiveConfig Web Application Installer (LC WAI) Library
 * @author Christoph Russow
 * @copyright Copyright (c) 2009-2024 LiveConfig GmbH.
 * @version 1.0.0
 * @package LC_WAI
 * --------------------------------------------------------------------------
 */

/**
 * API version code (to check API compatibility)
 */
define('WAI_API_VERSION', '1.0.6');

if(!defined('WAI_INCLUDE')) { die("ERROR direct access not allowed\n"); }

if(!class_exists("Installer")) {
  /**
   * Installer class
   */
  class Installer {

    /**
     * return all required configuration variables (except MySQL data)
     */
    public function wai_getvars($REQUIRED_USER_VARS, $FALLBACK_LANG = 'en') {
      if (($vars = $this->get_env_vars(array('LC_LANG'))) === false) {
        return;
      }

      print "OK\n";

      if (is_array($REQUIRED_USER_VARS) && !empty($REQUIRED_USER_VARS)) {
        foreach($REQUIRED_USER_VARS as $variable) {
          print $variable['name']
              . "\t" . $variable['type']
              . "\t" . $variable['regex']
              . "\t" . (isset($variable['title'][$vars['LC_LANG']]) ? $variable['title'][$vars['LC_LANG']] : $variable['title'][$FALLBACK_LANG])
              . "\t" . (isset($variable['description'][$vars['LC_LANG']]) ? $variable['description'][$vars['LC_LANG']] : $variable['description'][$FALLBACK_LANG])
              . "\t" . (isset($variable['defaultvalue'][$vars['LC_LANG']]) ? $variable['defaultvalue'][$vars['LC_LANG']] : $variable['defaultvalue'][$FALLBACK_LANG])
              . "\n";
        }
      }
    }

    /**
     * download the package archive and (if available) any additional language packs from the distributors' page
     */
    public function wai_download($LCWAI_DOWNLOADS, $FALLBACK_LANG = 'en') {

      if (($vars = $this->get_env_vars(array('LC_DST','LC_LANG'))) === false) {
        return;
      }

      if(!isset($LCWAI_DOWNLOADS['ALL']) && !isset($LCWAI_DOWNLOADS[$vars['LC_LANG']])) {
        $vars['LC_LANG'] = $FALLBACK_LANG;
      }

      foreach ($LCWAI_DOWNLOADS as $language => $lang_downloads) {
        if ($language == "ALL" || $language == $vars['LC_LANG']) {
          if (is_array($lang_downloads)) {
            foreach ($lang_downloads as $download) {
              // download from $download['URL'] to $vars['LC_DST'].'/'.$download['NAME'] and check sha1sum with $download['SHA1']
              if ($this->file_download($download['URL'], $vars['LC_DST'].'/'.$download['NAME'], $download['SHA1']) === false) {
                return false;
              }
            }
          }
        }
      }

      print "OK\n";
    }

    /**
     * update the package (security fixes/minor releases) if there are updates available
     */
    public function wai_update() {
      print "ERROR Update not implemented yet!\n";
    }

    /**
     * upgrade the package to next major release (if there is an update available)
     */
    public function wai_upgrade() {
      print "ERROR Upgrade not implemented yet!\n";
    }

    /**
     * do several preflight checks
     */
    public function wai_check($what) {
      $return_txt = "";

      if(isset($what['database']) && $what['database'] === true) {
        // Check Database credentials
        if (($vars = $this->get_env_vars(array('LC_MYSQL_HOST', 'LC_MYSQL_USER', 'LC_MYSQL_PW', 'LC_MYSQL_DB'))) === false) {
          return;
        }
        if(($dbh = mysql_connect($vars['LC_MYSQL_HOST'], $vars['LC_MYSQL_USER'], $vars['LC_MYSQL_PW'])) === false) {
          print "ERROR while connecting to database\n";
          return;
        }
        if(mysql_select_db($vars['LC_MYSQL_DB'], $dbh) === false) {
          print "ERROR while selecting database\n";
          return;
        }

        //check if tables exist in database
        if(($result = mysql_query("SHOW TABLES", $dbh)) === false) {
          print "ERROR while executing 'SHOW TABLES'\n";
          mysql_close($dbh);
          return;
        }
        if(($tblcnt = mysql_num_rows($result)) === false) {
          print "ERROR while fetching table count\n";
          mysql_close($dbh);
          return;
        }
        if($tblcnt > 0) {
          $return_txt .= "TAB_EXIST\t1\n";
        }

        mysql_close($dbh);
      }

      print "OK\n".$return_txt;
    }

    /*
     * ------------------------- common functions -------------------------
     */

    /**
     * create randomized security string
     */
    public function get_random_str($length, $special_chars = false) {
      if (!$special_chars) {
        $init = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789';
      } else {
        $init = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789!"§%&/()=?`´*-_.:,;@|<>{[]}^°';
      }
      return substr(str_shuffle( str_repeat($init, 5 + round($length/strlen($init), 0)) ), 0, $length);
    }

    /**
     * moves files from $from to $to
     */
    public function move($from, $to) {

      if (($mv_bin = trim(shell_exec('which mv'))) == "") {
        print "ERROR mv not found!\n";
        return false;
      }

      //if last character is an asterisk (*) do a readdir and execute copy for every found file and folder
      // reason: escapeshellarg() adds ' ' arround the argument and * is no longer a wildcard.
      if (substr($from, -1) == "*") {
        $dir = substr($from, 0, strpos($from, "*"));
        if (($dhd = opendir($dir)) === false) {
          print "ERROR failed to open dir!\n";
          return false;
        }

        while (($file = readdir($dhd)) !== false) {
          if ($file != "." && $file != "..") {
            $command = $mv_bin . " " . escapeshellarg($dir."/".$file) . " " . escapeshellarg($to)." 1>&2";
            if (system($command, $ret) === false) {
              print "ERROR failed to execute mv (exitcode: $ret)!\n";
              return false;
            }
          }
        }

      } else {
        $command = $mv_bin . " " . escapeshellarg($from) . " " . escapeshellarg($to)." 1>&2";
        if (system($command, $ret) === false) {
          print "ERROR failed to execute mv (exitcode: $ret)!\n";
          return false;
        }
      }

      return true;
    }

    /**
     * copies files from $from to $to
     */
    public function copy($from, $to) {

      if (($cp_bin = trim(shell_exec('which cp'))) == "") {
        print "ERROR cp not found!\n";
        return false;
      }

      //if last character is an asterisk (*) do a readdir and execute copy for every found file and folder
      // reason: escapeshellarg() adds ' ' arround the argument and * is no longer a wildcard.
      if (substr($from, -1) == "*") {
        $dir = substr($from, 0, strpos($from, "*"));
        if (($dhd = opendir($dir)) === false) {
          print "ERROR failed to open dir!\n";
          return false;
        }

        while (($file = readdir($dhd)) !== false) {
          if ($file != "." && $file != "..") {
            $command = $cp_bin . " -a " . escapeshellarg($dir."/".$file) . " " . escapeshellarg($to)." 1>&2";
            if (system($command, $ret) === false) {
              print "ERROR failed to execute cp (exitcode: $ret)!\n";
              return false;
            }
          }
        }

      } else {
        $command = $cp_bin . " -a " . escapeshellarg($from) . " " . escapeshellarg($to)." 1>&2";
        if (system($command, $ret) === false) {
          print "ERROR failed to execute cp (exitcode: $ret)!\n";
          return false;
        }
      }
    }

    /**
     * remove()
     *   removes $what from disk (always with -rf enabled!)
     */
    public function remove($what) {
      if (($rm_bin = trim(shell_exec('which rm'))) == "") {
        print "ERROR mv not found!\n";
        return false;
      }

      if (substr($what, -1) == "*") {
        $dir = substr($what, 0, strpos($what, "*"));
        if (($dhd = opendir($dir)) === false) {
          print "ERROR failed to open dir!\n";
          return false;
        }

        while (($file = readdir($dhd)) !== false) {
          if ($file != "." && $file != "..") {
            $command = $rm_bin . " -rf " . escapeshellarg($dir."/".$file)." 1>&2";
            if (system($command, $ret) === false) {
              print "ERROR failed to execute mv (exitcode: $ret)!\n";
              return false;
            }
          }
        }

      } else {
        $command = $rm_bin . " -rf " . escapeshellarg($what)." 1>&2";
        if (system($command, $ret) === false) {
          print "ERROR failed to execute mv (exitcode: $ret)!\n";
          return false;
        }
      }

      return true;
    }

    /**
     * download a file either with wget or with url_fopen
     */
    public function file_download($url, $dst_file, $sha1sum) {

      // check if already downloaded
      if (!is_file($dst_file)) {
        if (($wget_bin = trim(shell_exec('which wget'))) == "") {
          // download with fopen (need url_fopen)
          if (ini_get('allow_url_fopen') == 0) {
            print "ERROR no wget and no url_fopen available\n";
            return false;
          }

          if (($dst = fopen($dst_file, "w")) === false) {
            print "ERROR failed to open dst file!\n";
            return false;
          }
          if (($src = fopen($url, "r")) === false) {
            print "ERROR failed to open src file!\n";
            return false;
          }
          $max_size = 50*1024*1024; // max download size in bytes
          $dl_size = 0; // current downloaded size

          while (($tmp = fread($src, 4096)) && $dl_size < $max_size) {
            $dl_size += 4096;
            fwrite($dst, $tmp);
          }

          fclose($src);
          fclose($dst);
        } else {
          // use wget for download
          $command = $wget_bin." -nv --no-check-certificate -O ". $dst_file." ".escapeshellarg($url)." 1>&2";
          if (system($command, $ret) === false) {
            print "ERROR failure while executing wget (exitcode: ".$ret.")!\n";
            return false;
          }
          if($ret > 0) {
            print "ERROR wget exited with error code ".$ret."\n";
            return false;
          }
        }
        // adjust permissions (independent of umask) so every user may read this file
        chmod($dst_file, 0644);
      }
      // check file checksum
      if ($sha1sum != null && $sha1sum != sha1_file($dst_file)) {
        print "ERROR in download sha1 check failed!\n";
        return false;
      }
      return true;
    }

    /*
     * extract all contents of the given package to the destination directory
     */
    public function package_extract($dst_dir, $file) {

      if (preg_match('/(.tar.z|.tar.gz|.tar.bz|.tgz|.tbz|.tar.bz2|.tbz2|.tar.xz|.txz|.tar.lzma)$/i', $file)) {
        $type = "tar";
      } else if(preg_match('/.zip$/', $file)) {
        $type = "zip";
      } else {
        $type = "";
      }

      switch($type) {
        // --- tar ---
        case 'tar':
          if (($tar_bin = trim(shell_exec('which tar'))) == "") {
            print "ERROR tar not found!\n";
            return false;
          }

          $command = $tar_bin . " -xf " . escapeshellarg($file) . " -C " . escapeshellarg($dst_dir)." 1>&2";

          if (system($command, $ret) === false) {
            print "ERROR failed to execute tar!\n";
            return false;
          }

          if ($ret != 0) {
            print "ERROR failure while executing tar\n";
            return false;
          }
          break;
        // --- zip ---
        case 'zip':
          if (($unzip_bin = trim(shell_exec('which unzip'))) == "") {
            print "ERROR unzip not found!\n";
            return false;
          }

          if (!chdir($dst_dir)) {
            print "ERROR failed to change dir!\n";
            return false;
          }

          $command = $unzip_bin . " -o " . escapeshellarg($file)." 1>&2";
          if (system($command, $ret) === false) {
            print "ERROR failed to execute unzip!\n";
            return false;
          }

          if ($ret != 0) {
            print "ERROR failure while executing unzip\n";
            return false;
          }
          break;
        // --- err ---
        default:
          print "ERROR unknown type!\n";
          break;
      }
      return true;
    }

    /**
     * read environment variables specified by Liveconfig
     */
    public function get_env_vars($mandatory, $optional = null) {

      $variables = array();
      if (!is_array($mandatory)) {
        return false;
      }
      foreach ($mandatory as $varname) {
        if (($variables[$varname] = getenv($varname)) === false) {
          print "ERROR ".$varname." not set\n";
          return false;
        }
      }

      if (is_array($optional) && !empty($optional)) {
        foreach ($optional as $varname) {
          if (($variables[$varname] = getenv($varname)) === false) {
            //not set -> optional -> ignore
          }
        }
      }

      return $variables;
    }

    /**
     * recursive chmod()
     */
    public function chmod($path, $filemode, $dirmode, $recursive = false) {
      if ($recursive) {
        if (is_dir($path)) {
          if (!chmod($path, $dirmode)) {
            print "ERROR chmod failed!\n";
            return false;
          }

          $dh = opendir($path);
          while (($file = readdir($dh)) !== false) {
            if ($file != "." && $file != "..") {
              if (!$this->chmod($path.'/'.$file, $filemode, $dirmode, $recursive)) {
                return false;
              }
            }
          }
          closedir($dh);
        } else {
          if (!is_link($path)) {
            if (!chmod($path, $filemode)) {
              print "ERROR chmod failed!\n";
              return false;
            }
          }
        }
      } else {
        if (is_dir($path)) {
          if (!chmod($path, $dirmode)) {
            print "ERROR chmod failed!\n";
            return false;
          }
        } else {
          if (is_link($path)) {
            return true;
          }
          if (!is_link($path)) {
            if (!chmod($path, $filemode)) {
              print "ERROR chmod failed!\n";
              return false;
            }
          }
        }
      }
      return true;
    }


    /**
     * read sql file and split it by query (one query per array element)
     */
    public function split_sqlfile($filename, $remove_non_ansi_comments = false, $support_mysql_statement_comments = false, $remove_inline_singleline_comments = false) {
      $lines = file($filename);

      if($lines === false) {
        /* empty file */
        return false;
      }
      return $this->split_sqlarray($lines, $remove_non_ansi_comments, $support_mysql_statement_comments, $remove_inline_singleline_comments);
    }

    /**
     * split array with sql file lines by query (one query per array element)
     */
    public function split_sqlarray($sqldata, $remove_non_ansi_comments = false, $support_mysql_statement_comments = false, $remove_inline_singleline_comments = false) {

      /* Remove the -- comments and append all lines if $remove_non_ansi_comments is true also remove the non ansi # comments */
      $scriptfile = "";
      foreach($sqldata as $line) {
        $line = trim($line);
        if($remove_inline_singleline_comments) {
          // need to ignore inline comments on this two line adding blocks
          if($remove_non_ansi_comments) {
            if(!preg_match('/^--/', $line) && !preg_match('/^#/', $line) && !preg_match('/(.+)--.+/', $line)) {
              $scriptfile.=" ".$line;
            }
          } else {
            if(!preg_match('/^--/', $line) && !preg_match('/(.+)--.+/', $line)) {
              $scriptfile.=" ".$line;
            }
          }
          // now we remove inline comments and add the sql to the script
          if(preg_match('/(.+)--.+/', $line, $matches)) {
            $scriptfile .= " ".$matches[1];
          }
        } else {
          if($remove_non_ansi_comments) {
            if(!preg_match('/^--/', $line) && !preg_match('/^#/', $line)) {
              $scriptfile.=" ".$line;
            }
          } else {
            if(!preg_match('/^--/', $line)) {
              $scriptfile.=" ".$line;
            }
          }
        }
      }

      if(empty($scriptfile)) {
        return false;
      }

      // remove the /* */  comments
      if($support_mysql_statement_comments) {
        $scriptfile  = preg_replace("/\/\*[^!].+?\*\//", "", $scriptfile);
      } else {
        $scriptfile  = preg_replace("/\/\*.+?\*\//", "", $scriptfile);
      }

      /* split long line to one query per line array and return */
      $hochkomma = false;
      $anfuehrungszeichen = false;
      $backtick = false;
      $fronttick = false;
      $newstr = "";
      $query_array = array();

      for($i=0; $i < strlen($scriptfile); $i++) {
        switch($scriptfile[$i]) {
          case '\'':
            $hochkomma ? $hochkomma = false : $hochkomma = true;
            $newstr .= $scriptfile[$i];
            break;
          case '"':
            $anfuehrungszeichen ? $anfuehrungszeichen = false : $anfuehrungszeichen = true;
            $newstr .= $scriptfile[$i];
            break;
          case "`":
            $backtick ? $backtick = false : $backtick = true;
            $newstr .= $scriptfile[$i];
            break;
          case "´":
            $fronttick ? $fronttick = false : $fronttick = true;
            $newstr .= $scriptfile[$i];
            break;
          case "\\":
            $newstr .= $scriptfile[$i];
            $i++;
            $newstr .= $scriptfile[$i];
            break;
          case ";":
            if(!$hochkomma && !$anfuehrungszeichen && !$backtick && !$fronttick) {
              $query_array[] = $newstr;
              $newstr = "";
            } else {
              $newstr .= $scriptfile[$i];
            }
            break;
          default:
            $newstr .= $scriptfile[$i];
            break;
        }
      }
      /* trim leading whitespaces */
      array_walk($query_array, array('Installer', 'trim_value'));
      return  $query_array;
    }

    /**
     * Helper function for split_sqlfile to trim leading spaces on every query via array_walk()
     */
    private function trim_value(&$value) {
      /* Callback für array_walk to trim the array entries */
      $value = trim($value);
    }

  }
}

?>
